import logging
import os

import pytest

import salt.crypt as crypt
import salt.exceptions
from tests.support.mock import patch


@pytest.fixture
def key_data():
    return [
        "-----BEGIN PUBLIC KEY-----",
        "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoe5QSDYRWKyknbVyRrIj",
        "rm1ht5HgKzAVUber0x54+b/UgxTd1cqI6I+eDlx53LqZSH3G8Rd5cUh8LHoGedSa",
        "E62vEiLAjgXa+RdgcGiQpYS8+Z2RvQJ8oIcZgO+2AzgBRHboNWHTYRRmJXCd3dKs",
        "9tcwK6wxChR06HzGqaOTixAuQlegWbOTU+X4dXIbW7AnuQBt9MCib7SxHlscrqcS",
        "cBrRvq51YP6cxPm/rZJdBqZhVrlghBvIpa45NApP5PherGi4AbEGYte4l+gC+fOA",
        "osEBis1V27djPpIyQS4qk3XAPQg6CYQMDltHqA4Fdo0Nt7SMScxJhfH0r6zmBFAe",
        "BQIDAQAB",
        "-----END PUBLIC KEY-----",
    ]


@pytest.fixture
def minion_root(tmp_path):
    root = tmp_path / "root"
    root.mkdir()
    (root / "etc").mkdir()
    (root / "etc" / "salt").mkdir()
    (root / "etc" / "salt" / "pki").mkdir()
    yield root


@pytest.mark.parametrize("linesep", ["\r\n", "\r", "\n"])
def test__clean_key(key_data, linesep):
    tst_key = linesep.join(key_data)
    chk_key = "\n".join(key_data)
    assert crypt.clean_key(tst_key) == crypt.clean_key(chk_key)


@pytest.mark.parametrize("linesep", ["\r\n", "\r", "\n"])
def test__clean_key_mismatch(key_data, linesep):
    tst_key = linesep.join(key_data)
    tst_key = tst_key.replace("5", "4")
    chk_key = "\n".join(key_data)
    assert crypt.clean_key(tst_key) != crypt.clean_key(chk_key)


async def test_auth_aes_key_rotation(minion_root, io_loop, caplog):
    pki_dir = minion_root / "etc" / "salt" / "pki"
    opts = {
        "id": "minion",
        "__role": "minion",
        "pki_dir": str(pki_dir),
        "master_uri": "tcp://127.0.0.1:4505",
        "keysize": 4096,
        "acceptance_wait_time": 60,
        "acceptance_wait_time_max": 60,
    }
    crypt.gen_keys(pki_dir, "minion", opts["keysize"])
    credskey = (
        opts["pki_dir"],  # where the keys are stored
        opts["id"],  # minion ID
        opts["master_uri"],  # master ID
        str(os.path.getmtime(os.path.join(opts["pki_dir"], "minion.pem"))),
    )

    aes = crypt.Crypticle.generate_key_string()
    session = crypt.Crypticle.generate_key_string()

    auth = crypt.AsyncAuth(opts, io_loop)

    async def mock_sign_in(*args, **kwargs):
        return mock_sign_in.response

    mock_sign_in.response = {
        "enc": "pub",
        "aes": aes,
        "session": session,
    }
    auth.sign_in = mock_sign_in

    assert credskey not in auth.creds_map

    with caplog.at_level(logging.DEBUG):
        await auth.authenticate()

    assert "Got new master aes key" in caplog.text
    assert credskey in auth.creds_map
    assert auth.creds_map[credskey]["aes"] == aes
    assert auth.creds_map[credskey]["session"] == session

    aes1 = crypt.Crypticle.generate_key_string()

    mock_sign_in.response = {
        "enc": "pub",
        "aes": aes1,
        "session": session,
    }

    with caplog.at_level(logging.DEBUG):
        await auth.authenticate()

    assert "The master's aes key has changed" in caplog.text
    assert credskey in auth.creds_map
    assert auth.creds_map[credskey]["aes"] == aes1
    assert auth.creds_map[credskey]["session"] == session

    session1 = crypt.Crypticle.generate_key_string()
    mock_sign_in.response = {
        "enc": "pub",
        "aes": aes1,
        "session": session1,
    }

    with caplog.at_level(logging.DEBUG):
        await auth.authenticate()

    assert "The master's session key has changed" in caplog.text
    assert credskey in auth.creds_map
    assert auth.creds_map[credskey]["aes"] == aes1
    assert auth.creds_map[credskey]["session"] == session1


def test_sauth_aes_key_rotation(minion_root, io_loop, caplog):

    pki_dir = minion_root / "etc" / "salt" / "pki"
    opts = {
        "id": "minion",
        "__role": "minion",
        "pki_dir": str(pki_dir),
        "master_uri": "tcp://127.0.0.1:4505",
        "keysize": 4096,
        "acceptance_wait_time": 60,
        "acceptance_wait_time_max": 60,
    }
    crypt.gen_keys(pki_dir, "minion", opts["keysize"])

    aes = crypt.Crypticle.generate_key_string()
    session = crypt.Crypticle.generate_key_string()

    auth = crypt.SAuth(opts, io_loop)

    def mock_sign_in(*args, **kwargs):
        return mock_sign_in.response

    mock_sign_in.response = {
        "enc": "pub",
        "aes": aes,
        "session": session,
    }
    auth.sign_in = mock_sign_in

    assert auth._creds is None

    with caplog.at_level(logging.DEBUG):
        auth.authenticate()

    assert "Got new master aes key" in caplog.text
    assert isinstance(auth._creds, dict)
    assert auth._creds["aes"] == aes
    assert auth._creds["session"] == session

    aes1 = crypt.Crypticle.generate_key_string()

    mock_sign_in.response = {
        "enc": "pub",
        "aes": aes1,
        "session": session,
    }

    with caplog.at_level(logging.DEBUG):
        auth.authenticate()

    assert "The master's aes key has changed" in caplog.text
    assert isinstance(auth._creds, dict)
    assert auth._creds["aes"] == aes1
    assert auth._creds["session"] == session

    session1 = crypt.Crypticle.generate_key_string()
    mock_sign_in.response = {
        "enc": "pub",
        "aes": aes1,
        "session": session1,
    }

    with caplog.at_level(logging.DEBUG):
        auth.authenticate()

    assert "The master's session key has changed" in caplog.text
    assert isinstance(auth._creds, dict)
    assert auth._creds["aes"] == aes1
    assert auth._creds["session"] == session1


def test_get_key_with_evict_bad_key(tmp_path):
    key_path = tmp_path / "key"
    key_path.write_text("asdfasoiasdofaoiu0923jnoiausbd98sb9")
    with pytest.raises(salt.exceptions.InvalidKeyError):
        crypt._get_key_with_evict(str(key_path), 1, None)


def test_async_auth_cache_private_key(minion_root, io_loop):

    pki_dir = minion_root / "etc" / "salt" / "pki"
    opts = {
        "id": "minion",
        "__role": "minion",
        "pki_dir": str(pki_dir),
        "master_uri": "tcp://127.0.0.1:4505",
        "keysize": 4096,
        "acceptance_wait_time": 60,
        "acceptance_wait_time_max": 60,
    }

    auth = crypt.AsyncAuth(opts, io_loop)

    # The private key is cached.
    assert isinstance(auth._private_key, crypt.PrivateKey)

    # get_keys returns the cached instance
    _id = id(auth._private_key)
    assert _id == id(auth.get_keys())


def test_async_auth_cache_token(minion_root, io_loop):
    pki_dir = minion_root / "etc" / "salt" / "pki"
    opts = {
        "id": "minion",
        "__role": "minion",
        "pki_dir": str(pki_dir),
        "master_uri": "tcp://127.0.0.1:4505",
        "keysize": 4096,
        "acceptance_wait_time": 60,
        "acceptance_wait_time_max": 60,
    }

    auth = crypt.AsyncAuth(opts, io_loop)

    with patch("salt.crypt.private_encrypt") as moc:
        auth.gen_token("salt")
        auth.gen_token("salt")
        moc.assert_called_once()



"""
Transactional executor module

.. versionadded:: 3004

"""

import os

import salt.utils.path

# Functions that are mapped into an equivalent one in
# transactional_update module
DELEGATION_MAP = {
    "state.single": "transactional_update.single",
    "state.sls": "transactional_update.sls",
    "state.apply": "transactional_update.apply",
    "state.highstate": "transactional_update.highstate",
}

# By default, all modules and functions are executed outside the
# transaction.  The next two sets will enumerate the exceptions that
# will be routed to transactional_update.call()
DEFAULT_DELEGATED_MODULES = [
    "ansible",
    "cabal",
    "chef",
    "cmd",
    "composer",
    "cp",
    "cpan",
    "cyg",
    "file",
    "freeze",
    "nix",
    "npm",
    "pip",
    "pkg",
    "puppet",
    "pyenv",
    "rbenv",
    "scp",
]
DEFAULT_DELEGATED_FUNCTIONS = []


def __virtual__():
    if salt.utils.path.which("transactional-update"):
        return True
    else:
        return (False, "transactional_update executor requires a transactional system")


def execute(opts, data, func, args, kwargs):
    """Delegate into transactional_update module

    The ``transactional_update`` module support the execution of
    functions inside a transaction, as support apply a state (via
    ``apply``, ``sls``, ``single`` or ``highstate``).

    This execution module can be used to route some Salt modules and
    functions to be executed inside the transaction snapshot.

    Add this executor in the minion configuration file:

    .. code-block:: yaml

        module_executors:
          - transactional_update
          - direct_call

    Or use the command line parameter:

    .. code-block:: bash

        salt-call --module-executors='[transactional_update, direct_call]' test.version

    You can also schedule a reboot if needed:

    .. code-block:: bash

        salt-call --module-executors='[transactional_update]' state.sls stuff activate_transaction=True

    There are some configuration parameters supported:

    .. code-block:: yaml

       # Replace the list of default modules that all the functions
       # are delegated to `transactional_update.call()`
       delegated_modules: [cmd, pkg]

       # Replace the list of default functions that are delegated to
       # `transactional_update.call()`
       delegated_functions: [pip.install]

       # Expand the default list of modules
       add_delegated_modules: [ansible]

       # Expand the default list of functions
       add_delegated_functions: [file.copy]

    """
    inside_transaction = os.environ.get("TRANSACTIONAL_UPDATE")

    fun = data["fun"]
    module, _ = fun.split(".")

    delegated_modules = set(opts.get("delegated_modules", DEFAULT_DELEGATED_MODULES))
    delegated_functions = set(
        opts.get("delegated_functions", DEFAULT_DELEGATED_FUNCTIONS)
    )
    if "executor_opts" in data:
        delegated_modules |= set(data["executor_opts"].get("add_delegated_modules", []))
        delegated_functions |= set(
            data["executor_opts"].get("add_delegated_functions", [])
        )
    else:
        delegated_modules |= set(opts.get("add_delegated_modules", []))
        delegated_functions |= set(opts.get("add_delegated_functions", []))

    if fun in DELEGATION_MAP and not inside_transaction:
        result = __executors__["direct_call.execute"](
            opts, data, __salt__[DELEGATION_MAP[fun]], args, kwargs
        )
    elif (
        module in delegated_modules or fun in delegated_functions
    ) and not inside_transaction:
        result = __salt__["transactional_update.call"](fun, *args, **kwargs)
    else:
        result = __executors__["direct_call.execute"](opts, data, func, args, kwargs)

    return result



"""
Wrap the cp module allowing for managed ssh file transfers
"""

import logging
import os

import salt.client.ssh
import salt.utils.files
import salt.utils.stringutils
import salt.utils.templates
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def get_file(path, dest, saltenv="base", makedirs=False, template=None, gzip=None):
    """
    Send a file from the master to the location in specified

    .. note::

        gzip compression is not supported in the salt-ssh version of
        cp.get_file. The argument is only accepted for interface compatibility.
    """
    if gzip is not None:
        log.warning("The gzip argument to cp.get_file in salt-ssh is unsupported")

    if template is not None:
        (path, dest) = _render_filenames(path, dest, saltenv, template)

    src = __context__["fileclient"].cache_file(
        path, saltenv, cachedir=os.path.join("salt-ssh", __salt__.kwargs["id_"])
    )
    single = salt.client.ssh.Single(__opts__, "", **__salt__.kwargs)
    ret = single.shell.send(src, dest, makedirs)
    return not ret[2]


def get_dir(path, dest, saltenv="base"):
    """
    Transfer a directory down
    """
    src = __context__["fileclient"].cache_dir(
        path, saltenv, cachedir=os.path.join("salt-ssh", __salt__.kwargs["id_"])
    )
    src = " ".join(src)
    single = salt.client.ssh.Single(__opts__, "", **__salt__.kwargs)
    ret = single.shell.send(src, dest)
    return not ret[2]


def get_url(path, dest, saltenv="base"):
    """
    retrieve a URL
    """
    src = __context__["fileclient"].cache_file(
        path, saltenv, cachedir=os.path.join("salt-ssh", __salt__.kwargs["id_"])
    )
    single = salt.client.ssh.Single(__opts__, "", **__salt__.kwargs)
    ret = single.shell.send(src, dest)
    return not ret[2]


def list_states(saltenv="base"):
    """
    List all the available state modules in an environment
    """
    return __context__["fileclient"].list_states(saltenv)


def list_master(saltenv="base", prefix=""):
    """
    List all of the files stored on the master
    """
    return __context__["fileclient"].file_list(saltenv, prefix)


def list_master_dirs(saltenv="base", prefix=""):
    """
    List all of the directories stored on the master
    """
    return __context__["fileclient"].dir_list(saltenv, prefix)


def list_master_symlinks(saltenv="base", prefix=""):
    """
    List all of the symlinks stored on the master
    """
    return __context__["fileclient"].symlink_list(saltenv, prefix)


def _render_filenames(path, dest, saltenv, template):
    """
    Process markup in the :param:`path` and :param:`dest` variables (NOT the
    files under the paths they ultimately point to) according to the markup
    format provided by :param:`template`.
    """
    if not template:
        return (path, dest)

    # render the path as a template using path_template_engine as the engine
    if template not in salt.utils.templates.TEMPLATE_REGISTRY:
        raise CommandExecutionError(
            f"Attempted to render file paths with unavailable engine {template}"
        )

    kwargs = {}
    kwargs["salt"] = __salt__.value()
    kwargs["pillar"] = __pillar__.value()
    kwargs["grains"] = __grains__.value()
    kwargs["opts"] = __opts__
    kwargs["saltenv"] = saltenv

    def _render(contents):
        """
        Render :param:`contents` into a literal pathname by writing it to a
        temp file, rendering that file, and returning the result.
        """
        # write out path to temp file
        tmp_path_fn = salt.utils.files.mkstemp()
        with salt.utils.files.fopen(tmp_path_fn, "w+") as fp_:
            fp_.write(salt.utils.stringutils.to_str(contents))
        data = salt.utils.templates.TEMPLATE_REGISTRY[template](
            tmp_path_fn, to_str=True, **kwargs
        )
        salt.utils.files.safe_rm(tmp_path_fn)
        if not data["result"]:
            # Failed to render the template
            raise CommandExecutionError(
                "Failed to render file path with error: {}".format(data["data"])
            )
        else:
            return data["data"]

    path = _render(path)
    dest = _render(dest)
    return (path, dest)



import logging
import threading
import time
from concurrent.futures.thread import ThreadPoolExecutor

import pytest
from saltfactories.utils import random_string

import salt.channel.server
import salt.master
from tests.support.pytest.transport import PubServerChannelProcess

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.skip_on_fips_enabled_platform,
    pytest.mark.skip_on_spawning_platform(
        reason="These tests are currently broken on spawning platforms. Need to be rewritten.",
    ),
    pytest.mark.skipif(
        "grains['osfinger'] == 'Rocky Linux-8' and grains['osarch'] == 'aarch64'",
        reason="Temporarily skip on Rocky Linux 8 Arm64",
    ),
]


@pytest.fixture(scope="module", params=["tcp", "zeromq"])
def transport(request):
    yield request.param


@pytest.fixture(scope="module")
def salt_master(salt_factories, transport):
    config_defaults = {
        "transport": transport,
        "auto_accept": True,
        "sign_pub_messages": False,
    }
    factory = salt_factories.salt_master_daemon(
        random_string("ipc-master-"), defaults=config_defaults
    )
    return factory


@pytest.fixture(scope="module")
def salt_minion(salt_master):
    config_defaults = {
        "transport": salt_master.config["transport"],
        "master_ip": "127.0.0.1",
        "master_port": salt_master.config["ret_port"],
        "auth_timeout": 5,
        "auth_tries": 1,
        "master_uri": "tcp://127.0.0.1:{}".format(salt_master.config["ret_port"]),
    }
    factory = salt_master.salt_minion_daemon(
        random_string("zeromq-minion-"), defaults=config_defaults
    )
    return factory


@pytest.mark.skip_on_windows
@pytest.mark.slow_test
def test_publish_to_pubserv_ipc(salt_master, salt_minion, transport):
    """
    Test sending 10K messags to ZeroMQPubServerChannel using IPC transport

    ZMQ's ipc transport not supported on Windows
    """
    opts = dict(
        salt_master.config.copy(),
        ipc_mode="ipc",
        pub_hwm=0,
        transport=transport,
    )
    minion_opts = dict(salt_minion.config.copy(), transport=transport)
    with PubServerChannelProcess(opts, minion_opts) as server_channel:
        send_num = 10000
        expect = []
        for idx in range(send_num):
            expect.append(idx)
            load = {"tgt_type": "glob", "tgt": "*", "jid": idx}
            server_channel.publish(load)
    results = server_channel.collector.results
    assert len(results) == send_num, "{} != {}, difference: {}".format(
        len(results), send_num, set(expect).difference(results)
    )


@pytest.mark.skip_on_freebsd
@pytest.mark.slow_test
def test_issue_36469_tcp(salt_master, salt_minion, transport):
    """
    Test sending both large and small messags to publisher using TCP

    https://github.com/saltstack/salt/issues/36469
    """
    #  if transport == "tcp":
    #      pytest.skip("Test not applicable to the ZeroMQ transport.")

    def _send_small(queue, finished_event, sid, num=10):
        try:
            for idx in range(num):
                load = {"tgt_type": "glob", "tgt": "*", "jid": f"{sid}-s{idx}"}
                queue.put(load)
        finally:
            finished_event.set()

    def _send_large(queue, finished_event, sid, num=10, size=250000 * 3):
        try:
            for idx in range(num):
                load = {
                    "tgt_type": "glob",
                    "tgt": "*",
                    "jid": f"{sid}-l{idx}",
                    "xdata": "0" * size,
                }
                queue.put(load)
        finally:
            finished_event.set()

    opts = dict(salt_master.config.copy(), ipc_mode="tcp", pub_hwm=0)
    send_num = 10 * 4
    expect = []
    with PubServerChannelProcess(opts, salt_minion.config.copy()) as server_channel:
        assert "aes" in salt.master.SMaster.secrets
        finished_events = [threading.Event() for _ in range(4)]
        with ThreadPoolExecutor(max_workers=4) as executor:
            executor.submit(_send_small, server_channel.queue, finished_events[0], 1)
            executor.submit(_send_large, server_channel.queue, finished_events[1], 2)
            executor.submit(_send_small, server_channel.queue, finished_events[2], 3)
            executor.submit(_send_large, server_channel.queue, finished_events[3], 4)
        expect.extend([f"{sid}-s{idx}" for sid in (1, 3) for idx in range(10)])
        expect.extend([f"{sid}-l{idx}" for sid in (2, 4) for idx in range(10)])

        # Wait for all expected publishes to be observed before leaving the context.
        expected_set = set(expect)
        # The overall test has a 90s timeout; cap the wait so we fail fast if
        # pubs are stuck but still allow slower builders a chance to flush.
        deadline = time.monotonic() + 45
        interval = 0.5
        missing = expected_set
        while time.monotonic() < deadline:
            received = set(server_channel.collector.results)
            missing = expected_set - received
            if not missing:
                break
            log.debug(
                "Collector waiting for publishes; received=%s missing=%s",
                len(received),
                list(sorted(missing))[:5],
            )
            time.sleep(interval)
        else:
            pytest.fail(
                "Collector timed out waiting for publishes: "
                f"received={len(received)} expected={len(expected_set)} "
                f"missing={list(sorted(missing))[:5]}"
            )
        for event in finished_events:
            if not event.wait(timeout=5):
                pytest.fail("Publisher thread did not signal completion")
    results = server_channel.collector.results
    assert len(results) == send_num, "{} != {}, difference: {}".format(
        len(results), send_num, set(expect).difference(results)
    )



import os

import pytest

from tests.support.case import ShellCase, SPMCase


@pytest.mark.windows_whitelisted
class SPMTest(ShellCase, SPMCase):
    """
    Test spm script
    """

    @pytest.mark.slow_test
    def test_spm_help(self):
        """
        test --help argument for spm
        """
        expected_args = ["--version", "--assume-yes", "--help"]
        output = self.run_spm("--help")
        for arg in expected_args:
            self.assertIn(arg, "".join(output))

    @pytest.mark.slow_test
    def test_spm_bad_arg(self):
        """
        test correct output when bad argument passed
        """
        expected_args = ["--version", "--assume-yes", "--help"]
        output = self.run_spm("doesnotexist")
        for arg in expected_args:
            self.assertIn(arg, "".join(output))

    @pytest.mark.slow_test
    def test_spm_assume_yes(self):
        """
        test spm install with -y arg
        """
        config = self._spm_config(assume_yes=False)
        self._spm_build_files(config)

        spm_file = os.path.join(config["spm_build_dir"], "apache-201506-2.spm")

        build = self.run_spm(f"build {self.formula_dir} -c {self._tmp_spm}")

        install = self.run_spm(f"install {spm_file} -c {self._tmp_spm} -y")

        self.assertTrue(
            os.path.exists(os.path.join(config["formula_path"], "apache", "apache.sls"))
        )

    @pytest.mark.slow_test
    def test_spm_force(self):
        """
        test spm install with -f arg
        """
        config = self._spm_config(assume_yes=False)
        self._spm_build_files(config)

        spm_file = os.path.join(config["spm_build_dir"], "apache-201506-2.spm")

        build = self.run_spm(f"build {self.formula_dir} -c {self._tmp_spm}")

        install = self.run_spm(f"install {spm_file} -c {self._tmp_spm} -y")

        self.assertTrue(
            os.path.exists(os.path.join(config["formula_path"], "apache", "apache.sls"))
        )

        # check if it forces the install after its already been installed it
        install = self.run_spm(f"install {spm_file} -c {self._tmp_spm} -y -f")

        self.assertEqual(["... installing apache"], install)



from collections import OrderedDict

import salt.modules.jboss7 as jboss7
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock
from tests.support.unit import TestCase


class JBoss7TestCase(TestCase, LoaderModuleMockMixin):

    jboss_config = {}
    org_run_operation = None

    def setup_loader_modules(self):
        self.org_run_operation = MagicMock()
        self.addCleanup(delattr, self, "org_run_operation")
        return {
            jboss7: {"__salt__": {"jboss7_cli.run_operation": self.org_run_operation}}
        }

    def test_create_simple_binding(self):
        jboss7.create_simple_binding(self.jboss_config, "java:global/env", "DEV")

        self.org_run_operation.assert_called_with(
            self.jboss_config,
            '/subsystem=naming/binding="java:global/env":add(binding-type=simple,'
            ' value="DEV")',
        )

    def test_create_simple_binding_with_backslash(self):
        jboss7.create_simple_binding(self.jboss_config, "java:global/env", r"DEV\2")

        self.org_run_operation.assert_called_with(
            self.jboss_config,
            r'/subsystem=naming/binding="java:global/env":add(binding-type=simple,'
            r' value="DEV\\\\2")',
        )

    def test_update_binding(self):
        jboss7.update_simple_binding(self.jboss_config, "java:global/env", "INT")

        self.org_run_operation.assert_called_with(
            self.jboss_config,
            '/subsystem=naming/binding="java:global/env":write-attribute(name=value,'
            ' value="INT")',
        )

    def test_update_binding_with_backslash(self):
        jboss7.update_simple_binding(self.jboss_config, "java:global/env", r"INT\2")

        self.org_run_operation.assert_called_with(
            self.jboss_config,
            r'/subsystem=naming/binding="java:global/env":write-attribute(name=value,'
            r' value="INT\\\\2")',
        )

    def test_read_binding(self):
        def cli_command_response(jboss_config, cli_command):
            if (
                cli_command
                == '/subsystem=naming/binding="java:global/env":read-resource'
            ):
                return {
                    "outcome": "success",
                    "result": {"binding-type": "simple", "value": "DEV"},
                }

        self.org_run_operation.side_effect = cli_command_response

        result = jboss7.read_simple_binding(self.jboss_config, "java:global/env")
        self.assertEqual(result["outcome"], "success")
        self.assertEqual(result["result"]["value"], "DEV")

    def test_create_datasource_all_properties_included(self):
        def cli_command_response(jboss_config, cli_command, fail_on_error=False):
            if (
                cli_command
                == '/subsystem=datasources/data-source="appDS":read-resource-description'
            ):
                return {
                    "outcome": "success",
                    "result": {
                        "attributes": {
                            "driver-name": {"type": "STRING"},
                            "connection-url": {"type": "STRING"},
                            "jndi-name": {"type": "STRING"},
                            "user-name": {"type": "STRING"},
                            "password": {"type": "STRING"},
                        }
                    },
                }

        self.org_run_operation.side_effect = cli_command_response

        datasource_properties = OrderedDict()
        datasource_properties["driver-name"] = "mysql"
        datasource_properties["connection-url"] = "jdbc:mysql://localhost:3306/app"
        datasource_properties["jndi-name"] = "java:jboss/datasources/appDS"
        datasource_properties["user-name"] = "app"
        datasource_properties["password"] = "app_password"

        jboss7.create_datasource(self.jboss_config, "appDS", datasource_properties)

        self.org_run_operation.assert_called_with(
            self.jboss_config,
            '/subsystem=datasources/data-source="appDS":add(driver-name="mysql",connection-url="jdbc:mysql://localhost:3306/app",jndi-name="java:jboss/datasources/appDS",user-name="app",password="app_password")',
            fail_on_error=False,
        )

    def test_create_datasource_format_boolean_value_when_string(self):
        def cli_command_response(jboss_config, cli_command, fail_on_error=False):
            if (
                cli_command
                == '/subsystem=datasources/data-source="appDS":read-resource-description'
            ):
                return {
                    "outcome": "success",
                    "result": {"attributes": {"use-ccm": {"type": "BOOLEAN"}}},
                }

        self.org_run_operation.side_effect = cli_command_response
        datasource_properties = OrderedDict()
        datasource_properties["use-ccm"] = "true"

        jboss7.create_datasource(self.jboss_config, "appDS", datasource_properties)

        self.org_run_operation.assert_called_with(
            self.jboss_config,
            '/subsystem=datasources/data-source="appDS":add(use-ccm=true)',
            fail_on_error=False,
        )

    def test_create_datasource_format_boolean_value_when_boolean(self):
        def cli_command_response(jboss_config, cli_command, fail_on_error=False):
            if (
                cli_command
                == '/subsystem=datasources/data-source="appDS":read-resource-description'
            ):
                return {
                    "outcome": "success",
                    "result": {"attributes": {"use-ccm": {"type": "BOOLEAN"}}},
                }

        self.org_run_operation.side_effect = cli_command_response
        datasource_properties = OrderedDict()
        datasource_properties["use-ccm"] = True

        jboss7.create_datasource(self.jboss_config, "appDS", datasource_properties)

        self.org_run_operation.assert_called_with(
            self.jboss_config,
            '/subsystem=datasources/data-source="appDS":add(use-ccm=true)',
            fail_on_error=False,
        )

    def test_create_datasource_format_int_value_when_int(self):
        def cli_command_response(jboss_config, cli_command, fail_on_error=False):
            if (
                cli_command
                == '/subsystem=datasources/data-source="appDS":read-resource-description'
            ):
                return {
                    "outcome": "success",
                    "result": {"attributes": {"min-pool-size": {"type": "INT"}}},
                }

        self.org_run_operation.side_effect = cli_command_response
        datasource_properties = OrderedDict()
        datasource_properties["min-pool-size"] = 15

        jboss7.create_datasource(self.jboss_config, "appDS", datasource_properties)

        self.org_run_operation.assert_called_with(
            self.jboss_config,
            '/subsystem=datasources/data-source="appDS":add(min-pool-size=15)',
            fail_on_error=False,
        )

    def test_create_datasource_format_int_value_when_string(self):
        def cli_command_response(jboss_config, cli_command, fail_on_error=False):
            if (
                cli_command
                == '/subsystem=datasources/data-source="appDS":read-resource-description'
            ):
                return {
                    "outcome": "success",
                    "result": {"attributes": {"min-pool-size": {"type": "INT"}}},
                }

        self.org_run_operation.side_effect = cli_command_response
        datasource_properties = OrderedDict()
        datasource_properties["min-pool-size"] = "15"

        jboss7.create_datasource(self.jboss_config, "appDS", datasource_properties)

        self.org_run_operation.assert_called_with(
            self.jboss_config,
            '/subsystem=datasources/data-source="appDS":add(min-pool-size=15)',
            fail_on_error=False,
        )

    def test_read_datasource(self):
        def cli_command_response(jboss_config, cli_command):
            if (
                cli_command
                == '/subsystem=datasources/data-source="appDS":read-resource'
            ):
                return {
                    "outcome": "success",
                    "result": {
                        "driver-name": "mysql",
                        "connection-url": "jdbc:mysql://localhost:3306/app",
                        "jndi-name": "java:jboss/datasources/appDS",
                        "user-name": "app",
                        "password": "app_password",
                    },
                }

        self.org_run_operation.side_effect = cli_command_response

        ds_result = jboss7.read_datasource(self.jboss_config, "appDS")
        ds_properties = ds_result["result"]

        self.assertEqual(ds_properties["driver-name"], "mysql")
        self.assertEqual(
            ds_properties["connection-url"], "jdbc:mysql://localhost:3306/app"
        )
        self.assertEqual(ds_properties["jndi-name"], "java:jboss/datasources/appDS")
        self.assertEqual(ds_properties["user-name"], "app")
        self.assertEqual(ds_properties["password"], "app_password")

    def test_update_datasource(self):
        datasource_properties = {
            "driver-name": "mysql",
            "connection-url": "jdbc:mysql://localhost:3306/app",
            "jndi-name": "java:jboss/datasources/appDS",
            "user-name": "newuser",
            "password": "app_password",
        }

        def cli_command_response(jboss_config, cli_command, fail_on_error=False):
            if (
                cli_command
                == '/subsystem=datasources/data-source="appDS":read-resource-description'
            ):
                return {
                    "outcome": "success",
                    "result": {
                        "attributes": {
                            "driver-name": {"type": "STRING"},
                            "connection-url": {"type": "STRING"},
                            "jndi-name": {"type": "STRING"},
                            "user-name": {"type": "STRING"},
                            "password": {"type": "STRING"},
                        }
                    },
                }

            elif (
                cli_command
                == '/subsystem=datasources/data-source="appDS":read-resource'
            ):
                return {
                    "outcome": "success",
                    "result": {
                        "driver-name": "mysql",
                        "connection-url": "jdbc:mysql://localhost:3306/app",
                        "jndi-name": "java:jboss/datasources/appDS",
                        "user-name": "app",
                        "password": "app_password",
                    },
                }

            elif (
                cli_command
                == '/subsystem=datasources/data-source="appDS":write-attribute(name="user-name",value="newuser")'
            ):
                return {"outcome": "success", "success": True}

        self.org_run_operation.side_effect = cli_command_response

        jboss7.update_datasource(self.jboss_config, "appDS", datasource_properties)

        self.org_run_operation.assert_any_call(
            self.jboss_config,
            '/subsystem=datasources/data-source="appDS":write-attribute(name="user-name",value="newuser")',
            fail_on_error=False,
        )



"""
Helper functions for transport components to handle message framing
"""

import salt.utils.msgpack


def frame_msg(body, header=None, raw_body=False):  # pylint: disable=unused-argument
    """
    Frame the given message with our wire protocol
    """
    framed_msg = {}
    if header is None:
        header = {}

    framed_msg["head"] = header
    framed_msg["body"] = body
    return salt.utils.msgpack.dumps(framed_msg)


def frame_msg_ipc(body, header=None, raw_body=False):  # pylint: disable=unused-argument
    """
    Frame the given message with our wire protocol for IPC

    For IPC, we don't need to be backwards compatible, so
    use the more efficient "use_bin_type=True" on Python 3.
    """
    framed_msg = {}
    if header is None:
        header = {}

    framed_msg["head"] = header
    framed_msg["body"] = body
    return salt.utils.msgpack.dumps(framed_msg, use_bin_type=True)


def _decode_embedded_list(src):
    """
    Convert enbedded bytes to strings if possible.
    List helper.
    """
    output = []
    for elem in src:
        if isinstance(elem, dict):
            elem = _decode_embedded_dict(elem)
        elif isinstance(elem, list):
            elem = _decode_embedded_list(elem)
        elif isinstance(elem, bytes):
            try:
                elem = elem.decode()
            except UnicodeError:
                pass
        output.append(elem)
    return output


def _decode_embedded_dict(src):
    """
    Convert enbedded bytes to strings if possible.
    Dict helper.
    """
    output = {}
    for key, val in src.items():
        if isinstance(val, dict):
            val = _decode_embedded_dict(val)
        elif isinstance(val, list):
            val = _decode_embedded_list(val)
        elif isinstance(val, bytes):
            try:
                val = val.decode()
            except UnicodeError:
                pass
        if isinstance(key, bytes):
            try:
                key = key.decode()
            except UnicodeError:
                pass
        output[key] = val
    return output


def decode_embedded_strs(src):
    """
    Convert enbedded bytes to strings if possible.
    This is necessary because Python 3 makes a distinction
    between these types.

    This wouldn't be needed if we used "use_bin_type=True" when encoding
    and "encoding='utf-8'" when decoding. Unfortunately, this would break
    backwards compatibility due to a change in wire protocol, so this less
    than ideal solution is used instead.
    """
    if isinstance(src, dict):
        return _decode_embedded_dict(src)
    elif isinstance(src, list):
        return _decode_embedded_list(src)
    elif isinstance(src, bytes):
        try:
            return src.decode()
        except UnicodeError:
            return src
    else:
        return src



"""
Tests for the Vault runner
"""

import logging
import shutil

import pytest

from tests.conftest import FIPS_TESTRUN

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.timeout_unless_on_windows(120),
]


@pytest.fixture(scope="class")
def pillar_state_tree(tmp_path_factory):
    _pillar_state_tree = tmp_path_factory.mktemp("pillar")
    try:
        yield _pillar_state_tree
    finally:
        shutil.rmtree(str(_pillar_state_tree), ignore_errors=True)


@pytest.fixture(scope="class")
def pillar_salt_master(salt_factories, pillar_state_tree):
    config_defaults = {
        "pillar_roots": {"base": [str(pillar_state_tree)]},
        "open_mode": True,
        "ext_pillar": [{"vault": "path=does/not/matter"}],
        "sdbvault": {
            "driver": "vault",
        },
        "vault": {
            "auth": {"method": "token", "token": "testsecret"},
            "policies": [
                "salt_minion",
                "salt_minion_{minion}",
                "salt_role_{pillar[roles]}",
                "salt_unsafe_{grains[foo]}",
            ],
            "policies_cache_time": 0,
            "url": "http://127.0.0.1:8200",
        },
        "minion_data_cache": False,
    }
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }
    factory = salt_factories.salt_master_daemon(
        "vault-policy-int-master-uncached",
        defaults=config_defaults,
        overrides=config_overrides,
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="class")
def pillar_caching_salt_master(salt_factories, pillar_state_tree):
    config_defaults = {
        "pillar_roots": {"base": [str(pillar_state_tree)]},
        "open_mode": True,
        "vault": {
            "auth": {"method": "token", "token": "testsecret"},
            "policies": [
                "salt_minion",
                "salt_minion_{minion}",
                "salt_role_{pillar[roles]}",
                "salt_unsafe_{grains[foo]}",
            ],
            "policies_cache_time": 0,
            "url": "http://127.0.0.1:8200",
        },
        "minion_data_cache": True,
    }
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }
    factory = salt_factories.salt_master_daemon(
        "vault-policy-int-master-cached",
        defaults=config_defaults,
        overrides=config_overrides,
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="class")
def pillar_salt_minion(pillar_salt_master):
    assert pillar_salt_master.is_running()
    factory = pillar_salt_master.salt_minion_daemon(
        "vault-policy-int-minion-uncached-1",
        defaults={"open_mode": True, "grains": {"foo": "bar"}},
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="class")
def pillar_caching_salt_minion(pillar_caching_salt_master):
    assert pillar_caching_salt_master.is_running()
    factory = pillar_caching_salt_master.salt_minion_daemon(
        "vault-policy-int-minion-cached-1",
        defaults={"open_mode": True, "grains": {"foo": "bar"}},
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="class")
def pillar_salt_run_cli(pillar_salt_master):
    return pillar_salt_master.salt_run_cli()


@pytest.fixture(scope="class")
def pillar_caching_salt_run_cli(pillar_caching_salt_master):
    return pillar_caching_salt_master.salt_run_cli()


@pytest.fixture(scope="class")
def pillar_salt_call_cli(pillar_salt_minion):
    return pillar_salt_minion.salt_call_cli()


@pytest.fixture(scope="class")
def pillar_caching_salt_call_cli(pillar_caching_salt_minion):
    return pillar_caching_salt_minion.salt_call_cli()


class TestVaultPillarPolicyTemplatesWithoutCache:
    @pytest.fixture(autouse=True)
    def pillar_policy_tree(
        self,
        pillar_salt_master,
        pillar_salt_minion,
    ):
        top_pillar_contents = """
        base:
          '{}':
            - roles
        """.format(
            pillar_salt_minion.id
        )
        roles_pillar_contents = """
        roles:
          - minion
          - web
        """
        top_file = pillar_salt_master.pillar_tree.base.temp_file(
            "top.sls", top_pillar_contents
        )
        roles_file = pillar_salt_master.pillar_tree.base.temp_file(
            "roles.sls", roles_pillar_contents
        )

        with top_file, roles_file:
            yield

    @pytest.fixture()
    def pillar_exe_loop(self, pillar_state_tree, pillar_salt_minion):
        top_file = """
        base:
          '{}':
            - roles
            - exe_loop
        """.format(
            pillar_salt_minion.id
        )
        exe_loop_pillar = r"""
        bar: {{ salt["vault.read_secret"]("does/not/matter") }}
        """
        top_tempfile = pytest.helpers.temp_file("top.sls", top_file, pillar_state_tree)
        exe_loop_tempfile = pytest.helpers.temp_file(
            "exe_loop.sls", exe_loop_pillar, pillar_state_tree
        )

        with top_tempfile, exe_loop_tempfile:
            yield

    @pytest.fixture()
    def pillar_sdb_loop(self, pillar_state_tree, pillar_salt_minion):
        top_file = """
        base:
          '{}':
            - roles
            - sdb_loop
        """.format(
            pillar_salt_minion.id
        )
        sdb_loop_pillar = r"""
        foo: {{ salt["sdb.get"]("sdb://sdbvault/does/not/matter/val") }}
        """
        top_tempfile = pytest.helpers.temp_file("top.sls", top_file, pillar_state_tree)
        sdb_loop_tempfile = pytest.helpers.temp_file(
            "sdb_loop.sls", sdb_loop_pillar, pillar_state_tree
        )

        with top_tempfile, sdb_loop_tempfile:
            yield

    @pytest.fixture(autouse=True)
    def minion_data_cache_absent(self, pillar_salt_run_cli, pillar_salt_minion):
        ret = pillar_salt_run_cli.run(
            "cache.flush", f"minions/{pillar_salt_minion.id}", "data"
        )
        assert ret.returncode == 0
        cached = pillar_salt_run_cli.run(
            "cache.fetch", f"minions/{pillar_salt_minion.id}", "data"
        )
        assert cached.returncode == 0
        assert not cached.data
        yield

    def test_show_policies(self, pillar_salt_run_cli, pillar_salt_minion):
        """
        Test that pillar data is refreshed correctly before rendering policies when necessary.
        This test includes the prevention of loop exceptions by the ext_pillar module
        This refresh does not include grains and pillar data targeted by these grains (unsafe anyways!).
        """
        ret = pillar_salt_run_cli.run(
            "vault.show_policies", pillar_salt_minion.id, expire=0
        )
        assert ret.data == [
            "salt_minion",
            f"salt_minion_{pillar_salt_minion.id}",
            "salt_role_minion",
            "salt_role_web",
        ]
        assert "Pillar render error: Failed to load ext_pillar vault" not in ret.stderr

    def test_show_policies_uncached_data_no_pillar_refresh(
        self, pillar_salt_run_cli, pillar_salt_minion
    ):
        """
        Test that the pillar is not refreshed when explicitly disabled
        """
        ret = pillar_salt_run_cli.run(
            "vault.show_policies", pillar_salt_minion.id, refresh_pillar=False, expire=0
        )
        assert ret.data == ["salt_minion", f"salt_minion_{pillar_salt_minion.id}"]

    @pytest.mark.skipif(
        FIPS_TESTRUN, reason="Signing with SHA1 not supported in FIPS mode."
    )
    def test_policy_compilation_prevents_loop_for_execution_module(
        self,
        pillar_salt_run_cli,
        pillar_salt_minion,
        pillar_exe_loop,
    ):
        """
        Test that the runner prevents a recursive cycle from happening
        """
        ret = pillar_salt_run_cli.run(
            "vault.show_policies", pillar_salt_minion.id, refresh_pillar=True, expire=0
        )
        assert ret.data == [
            "salt_minion",
            f"salt_minion_{pillar_salt_minion.id}",
            "salt_role_minion",
            "salt_role_web",
        ]
        assert "Pillar render error: Rendering SLS 'exe_loop' failed" in ret.stderr
        assert "Cyclic dependency detected while refreshing pillar" in ret.stderr

    @pytest.mark.skipif(
        FIPS_TESTRUN, reason="Signing with SHA1 not supported in FIPS mode."
    )
    def test_policy_compilation_prevents_loop_for_sdb_module(
        self,
        pillar_salt_run_cli,
        pillar_salt_minion,
        pillar_sdb_loop,
    ):
        """
        Test that the runner prevents a recursive cycle from happening
        """
        ret = pillar_salt_run_cli.run(
            "vault.show_policies", pillar_salt_minion.id, refresh_pillar=True, expire=0
        )
        assert ret.data == [
            "salt_minion",
            f"salt_minion_{pillar_salt_minion.id}",
            "salt_role_minion",
            "salt_role_web",
        ]
        assert "Pillar render error: Rendering SLS 'sdb_loop' failed" in ret.stderr
        assert "Cyclic dependency detected while refreshing pillar" in ret.stderr


class TestVaultPillarPolicyTemplatesWithCache:
    @pytest.fixture(autouse=True)
    def pillar_caching_policy_tree(
        self, pillar_caching_salt_master, pillar_caching_salt_minion
    ):
        top_pillar_contents = """
        base:
          '{}':
            - roles
        """.format(
            pillar_caching_salt_minion.id
        )
        roles_pillar_contents = """
        roles:
          - minion
          - web
        """
        top_file = pillar_caching_salt_master.pillar_tree.base.temp_file(
            "top.sls", top_pillar_contents
        )
        roles_file = pillar_caching_salt_master.pillar_tree.base.temp_file(
            "roles.sls", roles_pillar_contents
        )

        with top_file, roles_file:
            yield

    @pytest.fixture(autouse=True)
    def minion_data_cache_present(
        self,
        pillar_caching_salt_call_cli,
        pillar_caching_policy_tree,
    ):
        ret = pillar_caching_salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True
        yield

    @pytest.fixture(autouse=True)
    def minion_data_cache_outdated(
        self,
        minion_data_cache_present,
        pillar_caching_salt_run_cli,
        pillar_caching_salt_master,
        pillar_caching_salt_minion,
    ):
        roles_pillar_new_contents = """
        roles:
          - minion
          - web
          - fresh
        """
        roles_file = pillar_caching_salt_master.pillar_tree.base.temp_file(
            "roles.sls", roles_pillar_new_contents
        )

        cached = pillar_caching_salt_run_cli.run(
            "cache.fetch", f"minions/{pillar_caching_salt_minion.id}", "data"
        )
        assert cached.returncode == 0
        assert cached.data
        assert "pillar" in cached.data
        assert "grains" in cached.data
        assert "roles" in cached.data["pillar"]
        assert ["minion", "web"] == cached.data["pillar"]["roles"]
        with roles_file:
            yield

    def test_show_policies_cached_data_no_pillar_refresh(
        self,
        pillar_caching_salt_run_cli,
        pillar_caching_salt_minion,
    ):
        """
        Test that pillar data from cache is used when it is available
        """
        ret = pillar_caching_salt_run_cli.run(
            "vault.show_policies", pillar_caching_salt_minion.id, expire=0
        )
        assert ret.data == [
            "salt_minion",
            f"salt_minion_{pillar_caching_salt_minion.id}",
            "salt_role_minion",
            "salt_role_web",
            "salt_unsafe_bar",
        ]

    def test_show_policies_refresh_pillar(
        self,
        pillar_caching_salt_run_cli,
        pillar_caching_salt_minion,
    ):
        """
        Test that pillar data is always refreshed when requested.
        """
        ret = pillar_caching_salt_run_cli.run(
            "vault.show_policies",
            pillar_caching_salt_minion.id,
            refresh_pillar=True,
            expire=0,
        )
        assert ret.data == [
            "salt_minion",
            f"salt_minion_{pillar_caching_salt_minion.id}",
            "salt_role_minion",
            "salt_role_web",
            "salt_role_fresh",
            "salt_unsafe_bar",
        ]



"""
    Unit tests for the salt.modules.nacl module
"""

import pytest

import salt.utils.stringutils
from tests.support.mock import patch

pytest.importorskip("nacl.public")
pytest.importorskip("nacl.secret")

import salt.modules.nacl as nacl


@pytest.fixture
def configure_loader_modules(minion_opts):
    utils = salt.loader.utils(minion_opts)
    funcs = salt.loader.minion_mods(minion_opts, utils=utils)
    return {
        nacl: {
            "__opts__": minion_opts,
            "__utils__": utils,
            "__salt__": funcs,
        },
    }


@pytest.fixture
def test_keys():
    # Generate the keys
    ret = nacl.keygen()
    assert "pk" in ret
    assert "sk" in ret
    return ret["pk"], ret["sk"]


@pytest.fixture
def test_data():
    unencrypted_data = salt.utils.stringutils.to_bytes("hello")
    return unencrypted_data


def test_fips_mode():
    """
    Nacl module does not load when fips_mode is True
    """
    opts = {"fips_mode": True}
    with patch("salt.modules.nacl.__opts__", opts, create=True):
        ret = salt.modules.nacl.__virtual__()
        assert ret == (False, "nacl module not available in FIPS mode")


def test_keygen(test_keys):
    """
    Test keygen
    """
    test_pk, test_sk = test_keys
    assert len(test_pk) == 44
    assert len(test_sk) == 44


def test_enc_dec(test_data, test_keys):
    """
    Generate keys, encrypt, then decrypt.
    """
    # Encrypt with pk
    test_pk, test_sk = test_keys
    encrypted_data = nacl.enc(data=test_data, pk=test_pk)

    # Decrypt with sk
    decrypted_data = nacl.dec(data=encrypted_data, sk=test_sk)
    assert test_data == decrypted_data


def test_sealedbox_enc_dec(test_data, test_keys):
    """
    Generate keys, encrypt, then decrypt.
    """
    # Encrypt with pk
    test_pk, test_sk = test_keys
    encrypted_data = nacl.sealedbox_encrypt(data=test_data, pk=test_pk)

    # Decrypt with sk
    decrypted_data = nacl.sealedbox_decrypt(data=encrypted_data, sk=test_sk)

    assert test_data == decrypted_data


def test_secretbox_enc_dec(test_data, test_keys):
    """
    Generate keys, encrypt, then decrypt.
    """
    # Encrypt with sk
    test_pk, test_sk = test_keys
    encrypted_data = nacl.secretbox_encrypt(data=test_data, sk=test_sk)

    # Decrypt with sk
    decrypted_data = nacl.secretbox_decrypt(data=encrypted_data, sk=test_sk)

    assert test_data == decrypted_data



import copy
import functools
import importlib
import importlib.machinery
import importlib.util
import inspect
import logging
import os
import pathlib
import re
import sys
import tempfile
import threading
import time
import traceback
import types
from collections.abc import MutableMapping
from zipimport import zipimporter  # pylint: disable=no-name-in-module

import salt.config
import salt.defaults.events
import salt.defaults.exitcodes
import salt.loader.context
import salt.syspaths
import salt.utils.args
import salt.utils.context
import salt.utils.data
import salt.utils.dictupdate
import salt.utils.event
import salt.utils.files
import salt.utils.lazy
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.versions
from salt.utils.decorators import Depends

try:
    # Try the stdlib C extension first
    import _contextvars as contextvars
except ImportError:
    # Py<3.7
    import contextvars

log = logging.getLogger(__name__)

# pylint: disable=no-member
MODULE_KIND_SOURCE = 1
MODULE_KIND_COMPILED = 2
MODULE_KIND_EXTENSION = 3
MODULE_KIND_PKG_DIRECTORY = 5
SUFFIXES = []
for suffix in importlib.machinery.EXTENSION_SUFFIXES:
    SUFFIXES.append((suffix, "rb", MODULE_KIND_EXTENSION))
for suffix in importlib.machinery.SOURCE_SUFFIXES:
    SUFFIXES.append((suffix, "rb", MODULE_KIND_SOURCE))
for suffix in importlib.machinery.BYTECODE_SUFFIXES:
    SUFFIXES.append((suffix, "rb", MODULE_KIND_COMPILED))
MODULE_KIND_MAP = {
    MODULE_KIND_SOURCE: importlib.machinery.SourceFileLoader,
    MODULE_KIND_COMPILED: importlib.machinery.SourcelessFileLoader,
    MODULE_KIND_EXTENSION: importlib.machinery.ExtensionFileLoader,
}
# pylint: enable=no-member


SALT_BASE_PATH = pathlib.Path(salt.syspaths.INSTALL_DIR).resolve()
LOADED_BASE_NAME = "salt.loaded"
PY3_PRE_EXT = re.compile(r"\.cpython-{}{}(\.opt-[1-9])?".format(*sys.version_info[:2]))

# Will be set to pyximport module at runtime if cython is enabled in config.
pyximport = None


def _generate_module(name):
    if name in sys.modules:
        return

    code = "'''Salt loaded {} parent module'''".format(name.split(".")[-1])
    # ModuleType can't accept a unicode type on PY2
    module = types.ModuleType(str(name))
    exec(code, module.__dict__)
    sys.modules[name] = module


def _mod_type(module_path):
    if module_path.startswith(str(SALT_BASE_PATH)):
        return "int"
    return "ext"


class FilterDictWrapper(MutableMapping):
    """
    Create a dict which wraps another dict with a specific key suffix on get

    This is to replace "filter_load"
    """

    def __init__(self, d, suffix):
        self._dict = d
        self.suffix = suffix

    def __setitem__(self, key, val):
        self._dict[key] = val

    def __delitem__(self, key):
        del self._dict[key]

    def __getitem__(self, key):
        return self._dict[key + self.suffix]

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        for key in self._dict:
            if key.endswith(self.suffix):
                yield key.replace(self.suffix, "")


class LoadedFunc:
    """
    The functions loaded by LazyLoader instances using subscript notation
    'a[k]' will be wrapped with LoadedFunc.

      - Makes sure functions are called with the correct loader's context.
      - Provides access to a wrapped func's __global__ attribute

    :param func str: The function name to wrap
    :param LazyLoader loader: The loader instance to use in the context when the wrapped callable is called.
    """

    def __init__(self, name, loader):
        self.name = name
        self.loader = loader
        functools.update_wrapper(self, self.func)

    @property
    def func(self):
        return self.loader._dict[self.name]

    def __getattr__(self, name):
        return getattr(self.func, name)

    def __call__(self, *args, **kwargs):
        run_func = self.func
        mod = sys.modules[run_func.__module__]
        # All modules we've imported should have __opts__ defined. There are
        # cases in the test suite where mod ends up being something other than
        # a module we've loaded.
        set_test = False
        if hasattr(mod, "__opts__"):
            if not isinstance(mod.__opts__, salt.loader.context.NamedLoaderContext):
                if "test" in self.loader.opts:
                    if self.loader.opts["test"] is False:
                        mod.__opts__["test"] = False
                    else:
                        mod.__opts__["test"] = True
                    set_test = True
        if self.loader.inject_globals:
            run_func = global_injector_decorator(self.loader.inject_globals)(run_func)
        ret = self.loader.run(run_func, *args, **kwargs)
        if set_test:
            self.loader.opts["test"] = mod.__opts__["test"]
        return ret

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r}>"


class LoadedMod:
    """
    This class is used as a proxy to a loaded module
    """

    __slots__ = ("mod", "loader")

    def __init__(self, mod, loader):
        """
        Return the wrapped func's globals via this object's __globals__
        attribute.
        """
        self.mod = mod
        self.loader = loader

    def __getattr__(self, name):
        """
        Run the wrapped function in the loader's context.
        """
        try:
            return self.loader[f"{self.mod}.{name}"]
        except KeyError:
            raise AttributeError(
                f"No attribute by the name of {name} was found on {self.mod}"
            )

    def __repr__(self):
        return "<{} module='{}.{}'>".format(
            self.__class__.__name__, self.loader.loaded_base_name, self.mod
        )


class LazyLoader(salt.utils.lazy.LazyDict):
    """
    A pseduo-dictionary which has a set of keys which are the
    name of the module and function, delimited by a dot. When
    the value of the key is accessed, the function is then loaded
    from disk and into memory.

    .. note::

        Iterating over keys will cause all modules to be loaded.

    :param list module_dirs: A list of directories on disk to search for modules
    :param dict opts: The salt options dictionary.
    :param str tag: The tag for the type of module to load
    :param func mod_type_check: A function which can be used to verify files
    :param dict pack: A dictionary of function to be packed into modules as they are loaded
    :param list whitelist: A list of modules to whitelist
    :param bool virtual_enable: Whether or not to respect the __virtual__ function when loading modules.
    :param str virtual_funcs: The name of additional functions in the module to call to verify its functionality.
                                If not true, the module will not load.
    :param list extra_module_dirs: A list of directories that will be able to import from
    :param str pack_self: Pack this module into a variable by this name into modules loaded
    :returns: A LazyLoader object which functions as a dictionary. Keys are 'module.function' and values
    are function references themselves which are loaded on-demand.
    # TODO:
        - move modules_max_memory into here
        - singletons (per tag)
    """

    mod_dict_class = dict

    def __init__(
        self,
        module_dirs,
        opts=None,
        tag="module",
        loaded_base_name=None,
        mod_type_check=None,
        pack=None,
        whitelist=None,
        virtual_enable=True,
        static_modules=None,
        proxy=None,
        virtual_funcs=None,
        extra_module_dirs=None,
        pack_self=None,
        # Once we get rid of __utils__, the keyword argument bellow should be removed
        _only_pack_properly_namespaced_functions=True,
    ):  # pylint: disable=W0231
        """
        In pack, if any of the values are None they will be replaced with an
        empty context-specific dict
        """

        self.parent_loader = None
        self.inject_globals = {}
        self.pack = {} if pack is None else pack
        for i in self.pack:
            if isinstance(self.pack[i], salt.loader.context.NamedLoaderContext):
                self.pack[i] = self.pack[i].value()
        if opts is None:
            opts = {}
        opts = copy.deepcopy(opts)
        for i in ["pillar", "grains"]:
            if i in opts and isinstance(
                opts[i], salt.loader.context.NamedLoaderContext
            ):
                opts[i] = opts[i].value()
        threadsafety = not opts.get("multiprocessing")
        self.context_dict = salt.utils.context.ContextDict(threadsafe=threadsafety)
        self.opts = self.__prep_mod_opts(opts)
        self.pack_self = pack_self

        self.module_dirs = module_dirs
        self.tag = tag
        self._gc_finalizer = None
        self.loaded_base_name = loaded_base_name or LOADED_BASE_NAME
        self.mod_type_check = mod_type_check or _mod_type
        self._only_pack_properly_namespaced_functions = (
            _only_pack_properly_namespaced_functions
        )

        if "__context__" not in self.pack:
            self.pack["__context__"] = None

        for k, v in self.pack.items():
            if v is None:  # if the value of a pack is None, lets make an empty dict
                self.context_dict.setdefault(k, {})
                self.pack[k] = salt.utils.context.NamespacedDictWrapper(
                    self.context_dict, k
                )

        self.whitelist = whitelist
        self.virtual_enable = virtual_enable
        self.initial_load = True

        # names of modules that we don't have (errors, __virtual__, etc.)
        self.missing_modules = {}  # mapping of name -> error
        self.loaded_modules = set()
        self.loaded_files = set()  # TODO: just remove them from file_mapping?
        self.static_modules = static_modules if static_modules else []

        if virtual_funcs is None:
            virtual_funcs = []
        self.virtual_funcs = virtual_funcs

        self.extra_module_dirs = extra_module_dirs if extra_module_dirs else []
        self._clean_module_dirs = []

        self.disabled = set(
            self.opts.get(
                "disable_{}{}".format(self.tag, "" if self.tag[-1] == "s" else "s"),
                [],
            )
        )

        # A map of suffix to description for imp
        self.suffix_map = {}
        # A list to determine precedence of extensions
        # Prefer packages (directories) over modules (single files)!
        self.suffix_order = [""]
        for suffix, mode, kind in SUFFIXES:
            self.suffix_map[suffix] = (suffix, mode, kind)
            self.suffix_order.append(suffix)

        self._lock = threading.RLock()
        with self._lock:
            self._refresh_file_mapping()

        super().__init__()  # late init the lazy loader
        # create all of the import namespaces
        _generate_module(f"{self.loaded_base_name}.int")
        _generate_module(f"{self.loaded_base_name}.int.{tag}")
        _generate_module(f"{self.loaded_base_name}.ext")
        _generate_module(f"{self.loaded_base_name}.ext.{tag}")

    def clean_modules(self):
        """
        Clean modules
        """
        for name in list(sys.modules):
            if name.startswith(self.loaded_base_name):
                del sys.modules[name]

    def __getitem__(self, item):
        """
        Override the __getitem__ in order to decorate the returned function if we need
        to last-minute inject globals
        """
        super().__getitem__(item)  # try to get the item from the dictionary
        return LoadedFunc(item, self)

    def __getattr__(self, mod_name):
        """
        Allow for "direct" attribute access-- this allows jinja templates to
        access things like `salt.test.ping()`
        """
        if mod_name in ("__getstate__", "__setstate__"):
            return object.__getattribute__(self, mod_name)

        # if we have an attribute named that, lets return it.
        try:
            return object.__getattr__(self, mod_name)  # pylint: disable=no-member
        except AttributeError:
            pass

        # otherwise we assume its jinja template access
        if mod_name not in self.loaded_modules and not self.loaded:
            for name in self._iter_files(mod_name):
                if name in self.loaded_files:
                    continue
                # if we got what we wanted, we are done
                if self._load_module(name) and mod_name in self.loaded_modules:
                    break
        if mod_name in self.loaded_modules:
            return LoadedMod(mod_name, self)
        else:
            raise AttributeError(mod_name)

    def __repr__(self):
        return "<{} module='{}.{}'>".format(
            self.__class__.__name__, self.loaded_base_name, self.tag
        )

    def missing_fun_string(self, function_name):
        """
        Return the error string for a missing function.

        This can range from "not available' to "__virtual__" returned False
        """
        mod_name = function_name.split(".")[0]
        if mod_name in self.loaded_modules:
            return f"'{function_name}' is not available."
        else:
            try:
                reason = self.missing_modules[mod_name]
            except KeyError:
                return f"'{function_name}' is not available."
            else:
                if reason is not None:
                    return "'{}' __virtual__ returned False: {}".format(
                        mod_name, reason
                    )
                else:
                    return f"'{mod_name}' __virtual__ returned False"

    def _refresh_file_mapping(self):
        """
        refresh the mapping of the FS on disk
        """
        # map of suffix to description for imp
        if (
            self.opts.get("cython_enable", True) is True
            and ".pyx" not in self.suffix_map
        ):
            try:
                global pyximport
                pyximport = __import__("pyximport")  # pylint: disable=import-error
                pyximport.install()
                # add to suffix_map so file_mapping will pick it up
                self.suffix_map[".pyx"] = tuple()
                if ".pyx" not in self.suffix_order:
                    self.suffix_order.append(".pyx")
            except ImportError:
                log.info(
                    "Cython is enabled in the options but not present "
                    "in the system path. Skipping Cython modules."
                )
        # Allow for zipimport of modules
        if (
            self.opts.get("enable_zip_modules", True) is True
            and ".zip" not in self.suffix_map
        ):
            self.suffix_map[".zip"] = tuple()
            if ".zip" not in self.suffix_order:
                self.suffix_order.append(".zip")
        # allow for module dirs
        self.suffix_map[""] = ("", "", MODULE_KIND_PKG_DIRECTORY)

        # create mapping of filename (without suffix) to (path, suffix)
        # The files are added in order of priority; dict preserves insertion order.
        self.file_mapping = {}

        opt_match = []

        def _replace_pre_ext(obj):
            """
            Hack so we can get the optimization level that we replaced (if
            any) out of the re.sub call below. We use a list here because
            it is a persistent data structure that we will be able to
            access after re.sub is called.
            """
            opt_match.append(obj)
            return ""

        for mod_dir in self.module_dirs:
            try:
                # Make sure we have a sorted listdir in order to have
                # expectable override results
                files = sorted(x for x in os.listdir(mod_dir) if x != "__pycache__")
            except OSError:
                continue  # Next mod_dir
            try:
                pycache_files = [
                    os.path.join("__pycache__", x)
                    for x in sorted(os.listdir(os.path.join(mod_dir, "__pycache__")))
                ]
            except OSError:
                pass
            else:
                files.extend(pycache_files)

            for filename in files:
                try:
                    dirname, basename = os.path.split(filename)
                    if basename.startswith("_"):
                        # skip private modules
                        # log messages omitted for obviousness
                        continue  # Next filename
                    f_noext, ext = os.path.splitext(basename)
                    f_noext = PY3_PRE_EXT.sub(_replace_pre_ext, f_noext)
                    try:
                        opt_level = int(opt_match.pop().group(1).rsplit("-", 1)[-1])
                    except (AttributeError, IndexError, ValueError):
                        # No regex match or no optimization level matched
                        opt_level = 0
                    try:
                        opt_index = self.opts["optimization_order"].index(opt_level)
                    except KeyError:
                        log.trace(
                            "Disallowed optimization level %d for module "
                            "name '%s', skipping. Add %d to the "
                            "'optimization_order' config option if you "
                            "do not want to ignore this optimization "
                            "level.",
                            opt_level,
                            f_noext,
                            opt_level,
                        )
                        continue

                    # make sure it is a suffix we support
                    if ext not in self.suffix_map:
                        continue  # Next filename
                    if f_noext in self.disabled:
                        log.trace(
                            "Skipping %s, it is disabled by configuration", filename
                        )
                        continue  # Next filename
                    fpath = os.path.join(mod_dir, filename)
                    # if its a directory, lets allow us to load that
                    if ext == "":
                        # is there something __init__?
                        subfiles = os.listdir(fpath)
                        for suffix in self.suffix_order:
                            if "" == suffix:
                                continue  # Next suffix (__init__ must have a suffix)
                            init_file = f"__init__{suffix}"
                            if init_file in subfiles:
                                break
                        else:
                            continue  # Next filename

                    try:
                        curr_ext = self.file_mapping[f_noext][1]
                        curr_opt_index = self.file_mapping[f_noext][2]
                    except KeyError:
                        pass
                    else:
                        if "" in (curr_ext, ext) and curr_ext != ext:
                            log.error(
                                "Module/package collision: '%s' and '%s'",
                                fpath,
                                self.file_mapping[f_noext][0],
                            )

                        if ext == ".pyc" and curr_ext == ".pyc":
                            # Check the optimization level
                            if opt_index >= curr_opt_index:
                                # Module name match, but a higher-priority
                                # optimization level was already matched, skipping.
                                continue
                        elif not curr_ext or self.suffix_order.index(
                            ext
                        ) >= self.suffix_order.index(curr_ext):
                            # Match found but a higher-priorty match already
                            # exists, so skip this.
                            continue

                    if not dirname and ext == ".pyc":
                        # On Python 3, we should only load .pyc files from the
                        # __pycache__ subdirectory (i.e. when dirname is not an
                        # empty string).
                        continue

                    # Made it this far - add it
                    self.file_mapping[f_noext] = (fpath, ext, opt_index)

                except OSError:
                    continue
        for smod in self.static_modules:
            f_noext = smod.split(".")[-1]
            self.file_mapping[f_noext] = (smod, ".o", 0)

    def clear(self):
        """
        Clear the dict
        """
        with self._lock:
            super().clear()  # clear the lazy loader
            self.loaded_files = set()
            self.missing_modules = {}
            self.loaded_modules = set()
            # if we have been loaded before, lets clear the file mapping since
            # we obviously want a re-do
            if hasattr(self, "opts"):
                self._refresh_file_mapping()
            self.initial_load = False

    def __prep_mod_opts(self, opts):
        """
        Strip out of the opts any logger instance
        """
        if "__grains__" not in self.pack:
            grains = opts.get("grains", {})
            if isinstance(grains, salt.loader.context.NamedLoaderContext):
                grains = grains.value()
            self.context_dict["grains"] = grains
            self.pack["__grains__"] = salt.utils.context.NamespacedDictWrapper(
                self.context_dict, "grains"
            )

        if "__pillar__" not in self.pack:
            pillar = opts.get("pillar", {})
            if isinstance(pillar, salt.loader.context.NamedLoaderContext):
                pillar = pillar.value()
            self.context_dict["pillar"] = pillar
            self.pack["__pillar__"] = salt.utils.context.NamespacedDictWrapper(
                self.context_dict, "pillar"
            )

        mod_opts = {}
        for key, val in list(opts.items()):
            if key == "logger":
                continue
            mod_opts[key] = val

        if "__opts__" not in self.pack:
            self.pack["__opts__"] = mod_opts

        return mod_opts

    def _iter_files(self, mod_name):
        """
        Iterate over all file_mapping files in order of closeness to mod_name
        """
        # do we have an exact match?
        if mod_name in self.file_mapping:
            yield mod_name

        # do we have a partial match?
        for k in self.file_mapping:
            if mod_name in k:
                yield k

        # anyone else? Bueller?
        for k in self.file_mapping:
            if mod_name not in k:
                yield k

    def _reload_submodules(self, mod):
        submodules = (
            getattr(mod, sname)
            for sname in dir(mod)
            if isinstance(getattr(mod, sname), mod.__class__)
        )

        # reload only custom "sub"modules
        for submodule in submodules:
            # it is a submodule if the name is in a namespace under mod
            if submodule.__name__.startswith(mod.__name__ + "."):
                importlib.reload(submodule)
                self._reload_submodules(submodule)

    def __populate_sys_path(self):
        for directory in self.extra_module_dirs:
            if directory not in sys.path:
                sys.path.append(directory)
                self._clean_module_dirs.append(directory)

    def __clean_sys_path(self):
        invalidate_path_importer_cache = False
        for directory in self._clean_module_dirs:
            if directory in sys.path:
                sys.path.remove(directory)
                invalidate_path_importer_cache = True
        self._clean_module_dirs = []

        # Be sure that sys.path_importer_cache do not contains any
        # invalid FileFinder references
        importlib.invalidate_caches()

        # Because we are mangling with importlib, we can find from
        # time to time an invalidation issue with
        # sys.path_importer_cache, that requires the removal of
        # FileFinder that remain None for the extra_module_dirs
        if invalidate_path_importer_cache:
            for directory in self.extra_module_dirs:
                if (
                    directory in sys.path_importer_cache
                    and sys.path_importer_cache[directory] is None
                ):
                    del sys.path_importer_cache[directory]

    def _load_module(self, name):
        mod = None
        fpath, suffix = self.file_mapping[name][:2]
        # if the fpath has `.cpython-3x` in it, but the running Py version
        # is 3.y, the following will cause us to return immediately and we won't try to import this .pyc.
        # This is for the unusual case where several Python versions share a single
        # source tree and drop their .pycs in the same __pycache__ folder.
        # If we were to load a .pyc for another Py version it's not a big problem
        # but the log will get spammed with "Bad Magic Number" messages that
        # can be very misleading if the user is debugging another problem.
        try:
            (implementation_tag, cache_tag_ver) = sys.implementation.cache_tag.split(
                "-"
            )
            if cache_tag_ver not in fpath and implementation_tag in fpath:
                log.trace(
                    "Trying to load %s on %s, returning False.",
                    fpath,
                    sys.implementation.cache_tag,
                )
                return False
        except AttributeError:
            # Most likely Py 2.7 or some other Python version we don't really support
            pass

        self.loaded_files.add(name)
        fpath_dirname = os.path.dirname(fpath)
        try:
            self.__populate_sys_path()
            sys.path.append(fpath_dirname)
            if suffix == ".pyx":
                mod = pyximport.load_module(name, fpath, tempfile.gettempdir())
            elif suffix == ".o":
                top_mod = __import__(fpath, globals(), locals(), [])
                comps = fpath.split(".")
                if len(comps) < 2:
                    mod = top_mod
                else:
                    mod = top_mod
                    for subname in comps[1:]:
                        mod = getattr(mod, subname)
            elif suffix == ".zip":
                mod = zipimporter(fpath).load_module(name)
            else:
                desc = self.suffix_map[suffix]
                # if it is a directory, we don't open a file
                try:
                    mod_namespace = ".".join(
                        (
                            self.loaded_base_name,
                            self.mod_type_check(fpath),
                            self.tag,
                            name,
                        )
                    )
                except TypeError:
                    mod_namespace = "{}.{}.{}.{}".format(
                        self.loaded_base_name,
                        self.mod_type_check(fpath),
                        self.tag,
                        name,
                    )
                if suffix == "":
                    # pylint: disable=no-member
                    # Package directory, look for __init__
                    loader_details = [
                        (
                            importlib.machinery.SourceFileLoader,
                            importlib.machinery.SOURCE_SUFFIXES,
                        ),
                        (
                            importlib.machinery.SourcelessFileLoader,
                            importlib.machinery.BYTECODE_SUFFIXES,
                        ),
                        (
                            importlib.machinery.ExtensionFileLoader,
                            importlib.machinery.EXTENSION_SUFFIXES,
                        ),
                    ]
                    file_finder = importlib.machinery.FileFinder(
                        fpath_dirname, *loader_details
                    )
                    spec = file_finder.find_spec(mod_namespace)
                    if spec is None:
                        raise ImportError()
                    # TODO: Get rid of load_module in favor of
                    # exec_module below. load_module is deprecated, but
                    # loading using exec_module has been causing odd things
                    # with the magic dunders we pack into the loaded
                    # modules, most notably with salt-ssh's __opts__.
                    mod = spec.loader.load_module()
                    # mod = importlib.util.module_from_spec(spec)
                    # spec.loader.exec_module(mod)
                    # pylint: enable=no-member
                    sys.modules[mod_namespace] = mod
                    # reload all submodules if necessary
                    if not self.initial_load:
                        self._reload_submodules(mod)
                else:
                    # pylint: disable=no-member
                    loader = MODULE_KIND_MAP[desc[2]](mod_namespace, fpath)
                    spec = importlib.util.spec_from_file_location(
                        mod_namespace, fpath, loader=loader
                    )
                    if spec is None:
                        raise ImportError()
                    # TODO: Get rid of load_module in favor of
                    # exec_module below. load_module is deprecated, but
                    # loading using exec_module has been causing odd things
                    # with the magic dunders we pack into the loaded
                    # modules, most notably with salt-ssh's __opts__.
                    mod = self.run(spec.loader.load_module)
                    # mod = importlib.util.module_from_spec(spec)
                    # spec.loader.exec_module(mod)
                    # pylint: enable=no-member
                    sys.modules[mod_namespace] = mod
        except OSError:
            raise
        except ImportError as exc:
            if "magic number" in str(exc):
                error_msg = (
                    "Failed to import {} {}. Bad magic number. If migrating from"
                    " Python2 to Python3, remove all .pyc files and try again.".format(
                        self.tag, name
                    )
                )
                log.warning(error_msg)
                self.missing_modules[name] = error_msg
            log.debug("Failed to import %s %s:\n", self.tag, name, exc_info=True)
            self.missing_modules[name] = exc
            return False
        except Exception as error:  # pylint: disable=broad-except
            log.error(
                "Failed to import %s %s, this is due most likely to a syntax error:\n",
                self.tag,
                name,
                exc_info=True,
            )
            self.missing_modules[name] = error
            return False
        except SystemExit as error:
            try:
                fn_, _, caller, _ = traceback.extract_tb(sys.exc_info()[2])[-1]
            except IndexError:
                pass
            else:
                tgt_fns = [
                    os.path.join("salt", "utils", "process.py"),
                    os.path.join("salt", "cli", "daemons.py"),
                    os.path.join("salt", "cli", "api.py"),
                ]
                for tgt_fn in tgt_fns:
                    if fn_.endswith(tgt_fn) and "_handle_signals" in caller:
                        # Race conditon, SIGTERM or SIGINT received while loader
                        # was in process of loading a module. Call sys.exit to
                        # ensure that the process is killed.
                        sys.exit(salt.defaults.exitcodes.EX_OK)
            log.error(
                "Failed to import %s %s as the module called exit()\n",
                self.tag,
                name,
                exc_info=True,
            )
            self.missing_modules[name] = error
            return False
        finally:
            sys.path.remove(fpath_dirname)
            self.__clean_sys_path()

        loader_context = salt.loader.context.LoaderContext()
        if hasattr(mod, "__salt_loader__"):
            if not isinstance(mod.__salt_loader__, salt.loader.context.LoaderContext):
                log.warning("Override  __salt_loader__: %s", mod)
                mod.__salt_loader__ = loader_context
        else:
            mod.__salt_loader__ = loader_context

        if hasattr(mod, "__opts__"):
            if not isinstance(mod.__opts__, salt.loader.context.NamedLoaderContext):
                if not hasattr(mod, "__orig_opts__"):
                    mod.__orig_opts__ = copy.deepcopy(mod.__opts__)
                mod.__opts__ = copy.deepcopy(mod.__orig_opts__)
                mod.__opts__.update(self.opts)
        else:
            if not hasattr(mod, "__orig_opts__"):
                mod.__orig_opts__ = {}
            mod.__opts__ = copy.deepcopy(mod.__orig_opts__)
            mod.__opts__.update(self.opts)

        # pack whatever other globals we were asked to
        for p_name, p_value in self.pack.items():
            if p_name == "__opts__":
                continue
            mod_named_context = getattr(mod, p_name, None)
            if hasattr(mod_named_context, "default"):
                default = copy.deepcopy(mod_named_context.default)
            else:
                default = None
            named_context = loader_context.named_context(p_name, default)
            if mod_named_context is None:
                setattr(mod, p_name, named_context)
            elif named_context != mod_named_context:
                log.debug("Override  %s: %s", p_name, mod)
                setattr(mod, p_name, named_context)
            else:
                setattr(mod, p_name, named_context)

        if self.pack_self is not None:
            mod_named_context = getattr(mod, self.pack_self, None)
            if hasattr(mod_named_context, "default"):
                default = copy.deepcopy(mod_named_context.default)
            else:
                default = None
            named_context = loader_context.named_context(self.pack_self, default)
            if mod_named_context is None:
                setattr(mod, self.pack_self, named_context)
            elif named_context != mod_named_context:
                log.debug("Override  %s: %s", self.pack_self, mod)
                setattr(mod, self.pack_self, named_context)
            else:
                setattr(mod, self.pack_self, named_context)

        module_name = mod.__name__.rsplit(".", 1)[-1]

        # Call a module's initialization method if it exists
        module_init = getattr(mod, "__init__", None)
        if inspect.isfunction(module_init):
            try:
                self.run(module_init, self.opts)
            except TypeError as e:
                log.error(e)
            except Exception:  # pylint: disable=broad-except
                err_string = "__init__ failed"
                log.debug(
                    "Error loading %s.%s: %s",
                    self.tag,
                    module_name,
                    err_string,
                    exc_info=True,
                )
                self.missing_modules[module_name] = err_string
                self.missing_modules[name] = err_string
                return False

        # if virtual modules are enabled, we need to look for the
        # __virtual__() function inside that module and run it.
        if self.virtual_enable:
            virtual_funcs_to_process = ["__virtual__"] + self.virtual_funcs
            for virtual_func in virtual_funcs_to_process:
                (
                    virtual_ret,
                    module_name,
                    virtual_err,
                    virtual_aliases,
                ) = self._process_virtual(mod, module_name, virtual_func)
                if virtual_err is not None:
                    log.trace(
                        "Error loading %s.%s: %s", self.tag, module_name, virtual_err
                    )

                # if _process_virtual returned a non-True value then we are
                # supposed to not process this module
                if virtual_ret is not True and module_name not in self.missing_modules:
                    # If a module has information about why it could not be loaded, record it
                    self.missing_modules[module_name] = virtual_err
                    self.missing_modules[name] = virtual_err
                    return False
        else:
            virtual_aliases = ()

        # If this is a proxy minion then MOST modules cannot work. Therefore, require that
        # any module that does work with salt-proxy-minion define __proxyenabled__ as a list
        # containing the names of the proxy types that the module supports.
        #
        # Render modules and state modules are OK though
        if "proxy" in self.opts:
            if self.tag in ["grains", "proxy"]:
                if not hasattr(mod, "__proxyenabled__") or (
                    self.opts["proxy"]["proxytype"] not in mod.__proxyenabled__
                    and "*" not in mod.__proxyenabled__
                ):
                    err_string = "not a proxy_minion enabled module"
                    self.missing_modules[module_name] = err_string
                    self.missing_modules[name] = err_string
                    return False

        try:
            funcs_to_load = mod.__load__
            log.debug(
                "The functions from module '%s' are being loaded from the "
                "provided __load__ attribute",
                module_name,
            )
        except AttributeError:
            try:
                funcs_to_load = mod.__all__
                log.debug(
                    "The functions from module '%s' are being loaded from the "
                    "provided __all__ attribute",
                    module_name,
                )
            except AttributeError:
                funcs_to_load = dir(mod)
                log.debug(
                    "The functions from module '%s' are being loaded by "
                    "dir() on the loaded module",
                    module_name,
                )

        # If we had another module by the same virtual name, we should put any
        # new functions under the existing dictionary.
        mod_names = [module_name] + list(virtual_aliases)

        for attr in funcs_to_load:
            if attr.startswith("_"):
                # private functions are skipped
                continue
            func = getattr(mod, attr)
            if not inspect.isfunction(func) and not isinstance(func, functools.partial):
                # Not a function!? Skip it!!!
                continue

            if (
                self._only_pack_properly_namespaced_functions
                and not func.__module__.startswith(self.loaded_base_name)
            ):
                # We're not interested in imported functions, only
                # functions defined(or namespaced) on the loaded module.
                continue

            # Let's get the function name.
            # If the module has the __func_alias__ attribute, it must be a
            # dictionary mapping in the form of(key -> value):
            #   <real-func-name> -> <desired-func-name>
            #
            # It default's of course to the found callable attribute name
            # if no alias is defined.
            funcname = getattr(mod, "__func_alias__", {}).get(attr, attr)
            for tgt_mod in mod_names:
                try:
                    full_funcname = ".".join((tgt_mod, funcname))
                except TypeError:
                    full_funcname = f"{tgt_mod}.{funcname}"
                # Save many references for lookups
                # Careful not to overwrite existing (higher priority) functions
                if full_funcname not in self._dict:
                    self._dict[full_funcname] = func
                    self._apply_outputter(func, mod)
                self.loaded_modules.add(tgt_mod)

        # enforce depends
        try:
            Depends.enforce_dependencies(self._dict, self.tag, name)
        except RuntimeError as exc:
            log.info(
                "Depends.enforce_dependencies() failed for the following reason: %s",
                exc,
            )

        return True

    def _load(self, key):
        """
        Load a single item if you have it
        """
        # if the key doesn't have a '.' then it isn't valid for this mod dict
        if not isinstance(key, str):
            raise KeyError("The key must be a string.")
        if "." not in key:
            raise KeyError(f"The key '{key}' should contain a '.'")
        mod_name, _ = key.split(".", 1)
        with self._lock:
            # It is possible that the key is in the dictionary after
            # acquiring the lock due to another thread loading it.
            if mod_name in self.missing_modules or key in self._dict:
                return True
            # if the modulename isn't in the whitelist, don't bother
            if self.whitelist and mod_name not in self.whitelist:
                log.error(
                    "Failed to load function %s because its module (%s) is "
                    "not in the whitelist: %s",
                    key,
                    mod_name,
                    self.whitelist,
                )
                raise KeyError(key)

            def _inner_load(mod_name):
                for name in self._iter_files(mod_name):
                    if name in self.loaded_files:
                        continue
                    # if we got what we wanted, we are done
                    if self._load_module(name) and key in self._dict:
                        return True
                return False

            # try to load the module
            ret = None
            reloaded = False
            # re-scan up to once, IOErrors or a failed load cause re-scans of the
            # filesystem
            while True:
                try:
                    ret = _inner_load(mod_name)
                    if not reloaded and ret is not True:
                        self._refresh_file_mapping()
                        reloaded = True
                        continue
                    break
                except OSError:
                    if not reloaded:
                        self._refresh_file_mapping()
                        reloaded = True
                    continue

        return ret

    def _load_all(self):
        """
        Load all of them
        """
        with self._lock:
            for name in self.file_mapping:
                if name in self.loaded_files or name in self.missing_modules:
                    continue
                self._load_module(name)

            self.loaded = True

    def reload_modules(self):
        with self._lock:
            self.loaded_files = set()
            self._load_all()

    def _apply_outputter(self, func, mod):
        """
        Apply the __outputter__ variable to the functions
        """
        if hasattr(mod, "__outputter__"):
            outp = mod.__outputter__
            if func.__name__ in outp:
                func.__outputter__ = outp[func.__name__]

    def _process_virtual(self, mod, module_name, virtual_func="__virtual__"):
        """
        Given a loaded module and its default name determine its virtual name

        This function returns a tuple. The first value will be either True or
        False and will indicate if the module should be loaded or not (i.e. if
        it threw and exception while processing its __virtual__ function). The
        second value is the determined virtual name, which may be the same as
        the value provided.

        The default name can be calculated as follows::

            module_name = mod.__name__.rsplit('.', 1)[-1]
        """

        # The __virtual__ function will return either a True or False value.
        # If it returns a True value it can also set a module level attribute
        # named __virtualname__ with the name that the module should be
        # referred to as.
        #
        # This allows us to have things like the pkg module working on all
        # platforms under the name 'pkg'. It also allows for modules like
        # augeas_cfg to be referred to as 'augeas', which would otherwise have
        # namespace collisions. And finally it allows modules to return False
        # if they are not intended to run on the given platform or are missing
        # dependencies.
        virtual_aliases = getattr(mod, "__virtual_aliases__", tuple())
        try:
            error_reason = None
            if hasattr(mod, "__virtual__") and inspect.isfunction(mod.__virtual__):
                try:
                    start = time.time()
                    virtual_attr = getattr(mod, virtual_func)
                    virtual = self.run(virtual_attr)
                    if isinstance(virtual, tuple):
                        error_reason = virtual[1]
                        virtual = virtual[0]
                    if self.opts.get("virtual_timer", False):
                        end = time.time() - start
                        msg = "Virtual function took {} seconds for {}".format(
                            end, module_name
                        )
                        log.warning(msg)
                except Exception as exc:  # pylint: disable=broad-except
                    error_reason = (
                        "Exception raised when processing __virtual__ function"
                        " for {}. Module will not be loaded: {}".format(
                            mod.__name__, exc
                        )
                    )
                    log.error(error_reason, exc_info_on_loglevel=logging.DEBUG)
                    virtual = None
                # Get the module's virtual name
                virtualname = getattr(mod, "__virtualname__", virtual)
                if not virtual:
                    # if __virtual__() evaluates to False then the module
                    # wasn't meant for this platform or it's not supposed to
                    # load for some other reason.

                    # Some modules might accidentally return None and are
                    # improperly loaded
                    if virtual is None:
                        log.warning(
                            "%s.__virtual__() is wrongly returning `None`. "
                            "It should either return `True`, `False` or a new "
                            "name. If you're the developer of the module "
                            "'%s', please fix this.",
                            mod.__name__,
                            module_name,
                        )

                    return (False, module_name, error_reason, virtual_aliases)

                # At this point, __virtual__ did not return a
                # boolean value, let's check for deprecated usage
                # or module renames
                if virtual is not True and module_name != virtual:
                    # The module is renaming itself. Updating the module name
                    # with the new name
                    log.trace("Loaded %s as virtual %s", module_name, virtual)

                    if virtualname != virtual:
                        # The __virtualname__ attribute does not match what's
                        # being returned by the __virtual__() function. This
                        # should be considered an error.
                        log.error(
                            "The module '%s' is showing some bad usage. Its "
                            "__virtualname__ attribute is set to '%s' yet the "
                            "__virtual__() function is returning '%s'. These "
                            "values should match!",
                            mod.__name__,
                            virtualname,
                            virtual,
                        )

                    module_name = virtualname

                # If the __virtual__ function returns True and __virtualname__
                # is set then use it
                elif virtual is True and virtualname != module_name:
                    if virtualname is not True:
                        module_name = virtualname

        except KeyError:
            # Key errors come out of the virtual function when passing
            # in incomplete grains sets, these can be safely ignored
            # and logged to debug, still, it includes the traceback to
            # help debugging.
            log.debug("KeyError when loading %s", module_name, exc_info=True)

        except Exception:  # pylint: disable=broad-except
            # If the module throws an exception during __virtual__()
            # then log the information and continue to the next.
            log.error(
                "Failed to read the virtual function for %s: %s",
                self.tag,
                module_name,
                exc_info=True,
            )
            return (False, module_name, error_reason, virtual_aliases)

        return (True, module_name, None, virtual_aliases)

    def run(self, _func_or_method, *args, **kwargs):
        """
        Run the `_func_or_method` in this loader's context
        """
        self._last_context = contextvars.copy_context()
        return self._last_context.run(self._run_as, _func_or_method, *args, **kwargs)

    def _run_as(self, _func_or_method, *args, **kwargs):
        """
        Handle setting up the context properly and call the method
        """
        self.parent_loader = None
        try:
            current_loader = salt.loader.context.loader_ctxvar.get()
        except LookupError:
            current_loader = None
        if current_loader is not self:
            self.parent_loader = current_loader
        token = salt.loader.context.loader_ctxvar.set(self)
        try:
            ret = _func_or_method(*args, **kwargs)
            if isinstance(ret, salt.loader.context.NamedLoaderContext):
                ret = ret.value()
            return ret
        finally:
            self.parent_loader = None
            salt.loader.context.loader_ctxvar.reset(token)

    def run_in_thread(self, _func_or_method, *args, **kwargs):
        """
        Run the function in a new thread with the context of this loader
        """
        argslist = [self, _func_or_method]
        argslist.extend(args)
        thread = threading.Thread(target=self.target, args=argslist, kwargs=kwargs)
        thread.start()
        return thread

    @staticmethod
    def target(loader, _func_or_method, *args, **kwargs):
        loader.run(_func_or_method, *args, **kwargs)


def global_injector_decorator(inject_globals):
    """
    Decorator used by the LazyLoader to inject globals into a function at
    execute time.

    globals
        Dictionary with global variables to inject
    """

    def inner_decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            with salt.utils.context.func_globals_inject(f, **inject_globals):
                return f(*args, **kwargs)

        return wrapper

    return inner_decorator



"""
    :codeauthor: Gareth J. Greenaway <gareth@saltstack.com>
"""

import copy
import logging

import pytest
from saltfactories.utils import random_string

import salt.proxy.ssh_sample as ssh_sample_proxy
from salt.utils.vt import TerminalException
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def proxy_minion_config_module(salt_master_factory):
    factory = salt_master_factory.salt_proxy_minion_daemon(
        random_string("proxy-minion-"),
    )
    return factory.config


@pytest.fixture
def proxy_minion_config(proxy_minion_config_module):

    minion_config = copy.deepcopy(proxy_minion_config_module)
    minion_config["proxy"]["proxytype"] = "ssh_sample"
    minion_config["proxy"]["host"] = "localhost"
    minion_config["proxy"]["username"] = "username"
    minion_config["proxy"]["password"] = "password"
    return minion_config


@pytest.fixture
def configure_loader_modules():
    return {ssh_sample_proxy: {}}


class MockSSHConnection:
    def __init__(self, *args, **kwargs):
        return None

    def sendline(self, *args, **kwargs):
        return "", ""


def test_init(proxy_minion_config):
    """
    check ssh_sample_proxy init method
    """

    with patch(
        "salt.utils.vt_helper.SSHConnection",
        MagicMock(autospec=True, return_value=MockSSHConnection()),
    ):
        with patch.dict(ssh_sample_proxy.__opts__, proxy_minion_config):
            ssh_sample_proxy.init(proxy_minion_config)
            assert "server" in ssh_sample_proxy.__context__
            assert "initialized" in ssh_sample_proxy.__context__
            assert ssh_sample_proxy.__context__["initialized"]


def test_initialized(proxy_minion_config):
    """
    check ssh_sample_proxy initialized method
    """

    with patch(
        "salt.utils.vt_helper.SSHConnection",
        MagicMock(autospec=True, return_value=MockSSHConnection()),
    ):
        with patch.dict(ssh_sample_proxy.__opts__, proxy_minion_config):
            ssh_sample_proxy.init(proxy_minion_config)

            ret = ssh_sample_proxy.initialized()
            assert ret


def test_grains(proxy_minion_config):
    """
    check ssh_sample_proxy grains method
    """

    GRAINS_INFO = """{
  "os": "SshExampleOS",
  "kernel": "0.0000001",
  "housecat": "Are you kidding?"
}
"""
    mock_sendline = MagicMock(autospec=True, side_effect=[("", ""), (GRAINS_INFO, "")])
    with patch.object(MockSSHConnection, "sendline", mock_sendline):
        with patch(
            "salt.utils.vt_helper.SSHConnection",
            MagicMock(autospec=True, return_value=MockSSHConnection()),
        ):
            with patch.dict(ssh_sample_proxy.__opts__, proxy_minion_config):
                ssh_sample_proxy.init(proxy_minion_config)

                ret = ssh_sample_proxy.grains()
                assert "os" in ret
                assert "kernel" in ret
                assert "housecat" in ret

                assert ret["os"] == "SshExampleOS"
                assert ret["kernel"] == "0.0000001"
                assert ret["housecat"] == "Are you kidding?"

    # Read from __context__ cache
    mock_sendline = MagicMock(autospec=True, side_effect=[("", ""), (GRAINS_INFO, "")])
    mock_context = {
        "grains_cache": {
            "os": "SSH-ExampleOS",
            "kernel": "0.0000002",
            "dog": "Not kidding.",
        }
    }

    with patch.object(MockSSHConnection, "sendline", mock_sendline):
        with patch(
            "salt.utils.vt_helper.SSHConnection",
            MagicMock(autospec=True, return_value=MockSSHConnection()),
        ):
            with patch.dict(ssh_sample_proxy.__opts__, proxy_minion_config):
                with patch.dict(ssh_sample_proxy.__context__, mock_context):
                    ssh_sample_proxy.init(proxy_minion_config)

                    ret = ssh_sample_proxy.grains()
                    assert "os" in ret
                    assert "kernel" in ret
                    assert "dog" in ret

                    assert ret["os"] == "SSH-ExampleOS"
                    assert ret["kernel"] == "0.0000002"
                    assert ret["dog"] == "Not kidding."


def test_grains_refresh(proxy_minion_config):
    """
    check ssh_sample_proxy grains_refresh method
    """

    GRAINS_INFO = """{
  "os": "SshExampleOS",
  "kernel": "0.0000001",
  "housecat": "Are you kidding?"
}
"""

    mock_sendline = MagicMock(autospec=True, side_effect=[("", ""), (GRAINS_INFO, "")])
    mock_context = {
        "grains_cache": {
            "os": "SSH-ExampleOS",
            "kernel": "0.0000002",
            "dog": "Not kidding.",
        }
    }

    with patch.object(MockSSHConnection, "sendline", mock_sendline):
        with patch(
            "salt.utils.vt_helper.SSHConnection",
            MagicMock(autospec=True, return_value=MockSSHConnection()),
        ):
            with patch.dict(ssh_sample_proxy.__opts__, proxy_minion_config):
                with patch.dict(ssh_sample_proxy.__context__, mock_context):
                    ssh_sample_proxy.init(proxy_minion_config)

                    ret = ssh_sample_proxy.grains_refresh()
                    assert "os" in ret
                    assert "kernel" in ret
                    assert "housecat" in ret

                    assert ret["os"] == "SshExampleOS"
                    assert ret["kernel"] == "0.0000001"
                    assert ret["housecat"] == "Are you kidding?"


def test_fns():
    """
    check ssh_sample_proxy fns method
    """

    ret = ssh_sample_proxy.fns()

    assert "details" in ret
    assert ret["details"] == (
        "This key is here because a function in "
        "grains/ssh_sample.py called fns() here in the proxymodule."
    )


def test_ping(proxy_minion_config):
    """
    check ssh_sample_proxy ping method
    """

    mock_sendline = MagicMock(autospec=True, side_effect=[("", ""), ("", "")])
    with patch.object(MockSSHConnection, "sendline", mock_sendline):
        with patch(
            "salt.utils.vt_helper.SSHConnection",
            MagicMock(autospec=True, return_value=MockSSHConnection()),
        ):
            with patch.dict(ssh_sample_proxy.__opts__, proxy_minion_config):
                ssh_sample_proxy.init(proxy_minion_config)

                ret = ssh_sample_proxy.ping()
                assert ret

    mock_sendline = MagicMock(autospec=True, side_effect=[("", ""), TerminalException])
    with patch.object(MockSSHConnection, "sendline", mock_sendline):
        with patch(
            "salt.utils.vt_helper.SSHConnection",
            MagicMock(autospec=True, return_value=MockSSHConnection()),
        ):
            with patch.dict(ssh_sample_proxy.__opts__, proxy_minion_config):
                ssh_sample_proxy.init(proxy_minion_config)
                ret = ssh_sample_proxy.ping()
                assert not ret


def test_package_list(proxy_minion_config):
    """
    check ssh_sample_proxy package_list method
    """

    PKG_LIST = """{
  "coreutils": "1.05"
}
"""

    mock_sendline = MagicMock(autospec=True, side_effect=[("", ""), (PKG_LIST, "")])
    with patch.object(MockSSHConnection, "sendline", mock_sendline):
        with patch(
            "salt.utils.vt_helper.SSHConnection",
            MagicMock(autospec=True, return_value=MockSSHConnection()),
        ):
            with patch.dict(ssh_sample_proxy.__opts__, proxy_minion_config):
                ssh_sample_proxy.init(proxy_minion_config)

                ret = ssh_sample_proxy.package_list()
                assert ret
                assert "coreutils" in ret
                assert ret["coreutils"] == "1.05"


def test_package_install(proxy_minion_config):
    """
    check ssh_sample_proxy package_list method
    """
    PKG_INSTALL = """{
  "redbull": "1.0"
}
"""

    mock_sendline = MagicMock(autospec=True, side_effect=[("", ""), (PKG_INSTALL, "")])
    with patch.object(MockSSHConnection, "sendline", mock_sendline):
        with patch(
            "salt.utils.vt_helper.SSHConnection",
            MagicMock(autospec=True, return_value=MockSSHConnection()),
        ):
            with patch.dict(ssh_sample_proxy.__opts__, proxy_minion_config):
                ssh_sample_proxy.init(proxy_minion_config)

                ret = ssh_sample_proxy.package_install("redbull")
                assert ret
                assert "redbull" in ret
                assert ret["redbull"] == "1.0"



import salt.utils.platform
from salt.exceptions import CommandExecutionError

try:
    import pywintypes
    import win32security
    import win32service

    HAS_WIN32 = True
except ImportError:
    HAS_WIN32 = False

SERVICE_TYPE = {
    1: "Kernel Driver",
    2: "File System Driver",
    4: "Adapter Driver",
    8: "Recognizer Driver",
    16: "Win32 Own Process",
    32: "Win32 Share Process",
    256: "Interactive",
    "kernel": 1,
    "filesystem": 2,
    "adapter": 4,
    "recognizer": 8,
    "own": 16,
    "share": 32,
}

SERVICE_CONTROLS = {
    1: "Stop",
    2: "Pause/Continue",
    4: "Shutdown",
    8: "Change Parameters",
    16: "Netbind Change",
    32: "Hardware Profile Change",
    64: "Power Event",
    128: "Session Change",
    256: "Pre-Shutdown",
    512: "Time Change",
    1024: "Trigger Event",
}

SERVICE_STATE = {
    1: "Stopped",
    2: "Start Pending",
    3: "Stop Pending",
    4: "Running",
    5: "Continue Pending",
    6: "Pause Pending",
    7: "Paused",
}

SERVICE_ERRORS = {0: "No Error", 1066: "Service Specific Error"}

SERVICE_START_TYPE = {
    "boot": 0,
    "system": 1,
    "auto": 2,
    "manual": 3,
    "disabled": 4,
    0: "Boot",
    1: "System",
    2: "Auto",
    3: "Manual",
    4: "Disabled",
}

SERVICE_ERROR_CONTROL = {
    0: "Ignore",
    1: "Normal",
    2: "Severe",
    3: "Critical",
    "ignore": 0,
    "normal": 1,
    "severe": 2,
    "critical": 3,
}

__virtualname__ = "win_service"


def __virtual__():
    """
    Only load if Win32 Libraries are installed
    """
    if not salt.utils.platform.is_windows():
        return False, "win_dacl: Requires Windows"

    if not HAS_WIN32:
        return False, "win_dacl: Requires pywin32"

    return __virtualname__


def info(name):
    """
    Get information about a service on the system

    Args:
        name (str): The name of the service. This is not the display name. Use
            ``get_service_name`` to find the service name.

    Returns:
        dict: A dictionary containing information about the service.

    CLI Example:

    .. code-block:: bash

        salt '*' service.info spooler
    """
    try:
        handle_scm = win32service.OpenSCManager(
            None, None, win32service.SC_MANAGER_CONNECT
        )
    except pywintypes.error as exc:
        raise CommandExecutionError(f"Failed to connect to the SCM: {exc.strerror}")

    try:
        handle_svc = win32service.OpenService(
            handle_scm,
            name,
            win32service.SERVICE_ENUMERATE_DEPENDENTS
            | win32service.SERVICE_INTERROGATE
            | win32service.SERVICE_QUERY_CONFIG
            | win32service.SERVICE_QUERY_STATUS,
        )
    except pywintypes.error as exc:
        raise CommandExecutionError(f"Failed To Open {name}: {exc.strerror}")

    try:
        config_info = win32service.QueryServiceConfig(handle_svc)
        status_info = win32service.QueryServiceStatusEx(handle_svc)

        try:
            description = win32service.QueryServiceConfig2(
                handle_svc, win32service.SERVICE_CONFIG_DESCRIPTION
            )
        except pywintypes.error:
            description = "Failed to get description"

        delayed_start = win32service.QueryServiceConfig2(
            handle_svc, win32service.SERVICE_CONFIG_DELAYED_AUTO_START_INFO
        )
    finally:
        win32service.CloseServiceHandle(handle_scm)
        win32service.CloseServiceHandle(handle_svc)

    ret = dict()
    try:
        sid = win32security.LookupAccountName("", f"NT Service\\{name}")[0]
        ret["sid"] = win32security.ConvertSidToStringSid(sid)
    except pywintypes.error:
        ret["sid"] = "Failed to get SID"

    ret["BinaryPath"] = config_info[3]
    ret["LoadOrderGroup"] = config_info[4]
    ret["TagID"] = config_info[5]
    ret["Dependencies"] = config_info[6]
    ret["ServiceAccount"] = config_info[7]
    ret["DisplayName"] = config_info[8]
    ret["Description"] = description
    ret["Status_ServiceCode"] = status_info["ServiceSpecificExitCode"]
    ret["Status_CheckPoint"] = status_info["CheckPoint"]
    ret["Status_WaitHint"] = status_info["WaitHint"]
    ret["StartTypeDelayed"] = delayed_start

    flags = list()
    for bit in SERVICE_TYPE:
        if isinstance(bit, int):
            if config_info[0] & bit:
                flags.append(SERVICE_TYPE[bit])

    ret["ServiceType"] = flags if flags else config_info[0]

    flags = list()
    for bit in SERVICE_CONTROLS:
        if status_info["ControlsAccepted"] & bit:
            flags.append(SERVICE_CONTROLS[bit])

    ret["ControlsAccepted"] = flags if flags else status_info["ControlsAccepted"]

    try:
        ret["Status_ExitCode"] = SERVICE_ERRORS[status_info["Win32ExitCode"]]
    except KeyError:
        ret["Status_ExitCode"] = status_info["Win32ExitCode"]

    try:
        ret["StartType"] = SERVICE_START_TYPE[config_info[1]]
    except KeyError:
        ret["StartType"] = config_info[1]

    try:
        ret["ErrorControl"] = SERVICE_ERROR_CONTROL[config_info[2]]
    except KeyError:
        ret["ErrorControl"] = config_info[2]

    try:
        ret["Status"] = SERVICE_STATE[status_info["CurrentState"]]
    except KeyError:
        ret["Status"] = status_info["CurrentState"]

    return ret



"""
Management of Redis server
==========================

.. versionadded:: 2014.7.0

:depends:   - redis Python module
:configuration: See :py:mod:`salt.modules.redis` for setup instructions.

.. code-block:: yaml

    key_in_redis:
      redis.string:
        - value: string data

The redis server information specified in the minion config file can be
overridden in states using the following arguments: ``host``, ``post``, ``db``,
``password``.

.. code-block:: yaml

    key_in_redis:
      redis.string:
        - value: string data
        - host: localhost
        - port: 6379
        - db: 0
        - password: somuchkittycat
"""

import copy

__virtualname__ = "redis"


def __virtual__():
    """
    Only load if the redis module is in __salt__
    """
    if "redis.set_key" in __salt__:
        return __virtualname__
    return (False, "redis module could not be loaded")


def string(name, value, expire=None, expireat=None, **connection_args):
    """
    Ensure that the key exists in redis with the value specified

    name
        Redis key to manage

    value
        Data to persist in key

    expire
        Sets time to live for key in seconds

    expireat
        Sets expiration time for key via UNIX timestamp, overrides `expire`
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Key already set to defined value",
    }

    old_key = __salt__["redis.get_key"](name, **connection_args)

    if old_key != value:
        __salt__["redis.set_key"](name, value, **connection_args)
        ret["changes"][name] = "Value updated"
        ret["comment"] = "Key updated to new value"

    if expireat:
        __salt__["redis.expireat"](name, expireat, **connection_args)
        ret["changes"]["expireat"] = f"Key expires at {expireat}"
    elif expire:
        __salt__["redis.expire"](name, expire, **connection_args)
        ret["changes"]["expire"] = f"TTL set to {expire} seconds"

    return ret


def absent(name, keys=None, **connection_args):
    """
    Ensure key absent from redis

    name
        Key to ensure absent from redis

    keys
        list of keys to ensure absent, name will be ignored if this is used
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Key(s) specified already absent",
    }

    if keys:
        if not isinstance(keys, list):
            ret["result"] = False
            ret["comment"] = "`keys` not formed as a list type"
            return ret
        delete_list = [
            key for key in keys if __salt__["redis.exists"](key, **connection_args)
        ]
        if not delete_list:
            return ret
        __salt__["redis.delete"](*delete_list, **connection_args)
        ret["changes"]["deleted"] = delete_list
        ret["comment"] = "Keys deleted"
        return ret

    if __salt__["redis.exists"](name, **connection_args):
        __salt__["redis.delete"](name, **connection_args)
        ret["comment"] = "Key deleted"
        ret["changes"]["deleted"] = [name]
    return ret


def slaveof(
    name,
    sentinel_host=None,
    sentinel_port=None,
    sentinel_password=None,
    **connection_args,
):
    """
    Set this redis instance as a slave.

    .. versionadded:: 2016.3.0

    name
        Master to make this a slave of

    sentinel_host
        Ip of the sentinel to check for the master

    sentinel_port
        Port of the sentinel to check for the master

    """
    ret = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "Failed to setup slave",
    }

    kwargs = copy.copy(connection_args)
    sentinel_master = __salt__["redis.sentinel_get_master_ip"](
        name, sentinel_host, sentinel_port, sentinel_password
    )
    if sentinel_master["master_host"] in __salt__["network.ip_addrs"]():
        ret["result"] = True
        ret["comment"] = f"Minion is the master: {name}"
        return ret

    first_master = __salt__["redis.get_master_ip"](**connection_args)
    if first_master == sentinel_master:
        ret["result"] = True
        ret["comment"] = f"Minion already slave of master: {name}"
        return ret

    if __opts__["test"] is True:
        ret["comment"] = "Minion will be made a slave of {}: {}".format(
            name, sentinel_master["host"]
        )
        ret["result"] = None
        return ret

    kwargs.update(**sentinel_master)
    __salt__["redis.slaveof"](**kwargs)

    current_master = __salt__["redis.get_master_ip"](**connection_args)
    if current_master != sentinel_master:
        return ret

    ret["result"] = True
    ret["changes"] = {
        "old": first_master,
        "new": current_master,
    }
    ret["comment"] = f"Minion successfully connected to master: {name}"

    return ret



import time

from saltfactories.utils import random_string


def test_event(event_listener, salt_cli, minion, master):
    event_tag = random_string("salt/test/event/")
    event_data = {"event.fire_master": random_string("syndic test event: ")}
    start_time = time.time()

    ret = salt_cli.run(
        "event.fire_master",
        tag=event_tag,
        data=event_data,
        minion_tgt="syndic",
        _timeout=15,
    )
    assert ret.data is True

    event_pattern = (master.id, event_tag)
    matched_events = event_listener.wait_for_events(
        [event_pattern], after_time=start_time, timeout=30
    )
    assert matched_events.found_all_events
    for event in matched_events:
        assert event.data["id"] == "syndic"
        assert event.data["cmd"] == "_minion_event"
        assert event.data["data"] == event_data



from types import SimpleNamespace

import pytest

from tests.support.mock import MagicMock, patch

pytest.importorskip("grp")

import grp

import salt.utils.user


def test_get_group_list():
    getpwname = SimpleNamespace(pw_gid=1000)
    getgrgid = MagicMock(side_effect=[SimpleNamespace(gr_name="remote")])
    group_lines = [
        ["games", "x", 50, []],
        ["salt", "x", 1000, []],
    ]
    getgrall = [grp.struct_group(comps) for comps in group_lines]
    with patch("os.getgrouplist", MagicMock(return_value=[50, 1000, 12000])), patch(
        "pwd.getpwnam", MagicMock(return_value=getpwname)
    ), patch("salt.utils.user._getgrall", MagicMock(return_value=getgrall)), patch(
        "grp.getgrgid", getgrgid
    ):
        group_list = salt.utils.user.get_group_list("salt")
        assert group_list == ["games", "remote", "salt"]
        getgrgid.assert_called_once()



import logging
import threading
import time

import pytest

import salt.modules.etcd_mod as etcd_mod
from salt.utils.etcd_util import get_conn
from tests.support.pytest.etcd import *  # pylint: disable=wildcard-import,unused-wildcard-import

docker = pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skip_if_binaries_missing("docker", "dockerd", check_all=False),
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        etcd_mod: {
            "__opts__": minion_opts,
            "__utils__": {
                "etcd_util.get_conn": get_conn,
            },
        },
    }


@pytest.fixture(scope="module")
def minion_config_overrides(etcd_profile):
    return etcd_profile


@pytest.fixture(scope="module")
def etcd_client(minion_opts, profile_name):
    return get_conn(minion_opts, profile=profile_name)


@pytest.fixture(scope="module")
def prefix():
    return "/salt/util/test"


@pytest.fixture(autouse=True)
def cleanup_prefixed_entries(etcd_client, prefix):
    """
    Cleanup after each test to ensure a consistent starting state.
    """
    try:
        assert etcd_client.get(prefix, recurse=True) is None
        yield
    finally:
        etcd_client.delete(prefix, recurse=True)


def test_basic_operations(subtests, profile_name, prefix):
    """
    Make sure we can do the basics
    """
    with subtests.test("There should be no entries at the start with our prefix."):
        assert etcd_mod.get_(prefix, recurse=True, profile=profile_name) is None

    with subtests.test("We should be able to set and retrieve simple values"):
        etcd_mod.set_(f"{prefix}/1", "one", profile=profile_name)
        assert (
            etcd_mod.get_(f"{prefix}/1", recurse=False, profile=profile_name) == "one"
        )

    with subtests.test("We should be able to update and retrieve those values"):
        updated = {
            "1": "not one",
            "2": {
                "3": "two-three",
                "4": "two-four",
            },
        }
        etcd_mod.update(updated, path=prefix, profile=profile_name)
        assert etcd_mod.get_(prefix, recurse=True, profile=profile_name) == updated

    with subtests.test("We should be list all top level values at a directory"):
        expected = {
            prefix: {
                f"{prefix}/1": "not one",
                f"{prefix}/2/": {},
            },
        }
        assert etcd_mod.ls_(path=prefix, profile=profile_name) == expected

    with subtests.test("We should be able to remove values and get a tree hierarchy"):
        updated = {
            "2": {
                "3": "two-three",
                "4": "two-four",
            },
        }
        etcd_mod.rm_(f"{prefix}/1", profile=profile_name)
        assert etcd_mod.tree(path=prefix, profile=profile_name) == updated

    with subtests.test("updates should be able to be caught by waiting in read"):
        return_list = []

        def wait_func(return_list):
            return_list.append(
                etcd_mod.watch(f"{prefix}/1", timeout=30, profile=profile_name)
            )

        wait_thread = threading.Thread(target=wait_func, args=(return_list,))
        wait_thread.start()
        time.sleep(1)
        etcd_mod.set_(f"{prefix}/1", "one", profile=profile_name)
        wait_thread.join()
        modified = return_list.pop()
        assert modified["key"] == f"{prefix}/1"
        assert modified["value"] == "one"


def test_with_missing_profile(subtests, prefix, etcd_version, etcd_port):
    """
    Test the correct response when the profile is missing and we can't connect
    """
    if etcd_version in (EtcdVersion.v2, EtcdVersion.v3_v2_mode) and etcd_port != 2379:
        # Only need to run this once
        with subtests.test("Test no profile and bad connection in get_"):
            assert etcd_mod.get_(f"{prefix}/1") is None

        with subtests.test("Test no profile and bad connection in set_"):
            assert etcd_mod.set_(f"{prefix}/1", "lol") is None

        with subtests.test("Test no profile and bad connection in update"):
            assert etcd_mod.update({f"{prefix}/1": "SIUUU"}) is None

        with subtests.test("Test no profile and bad connection in watch"):
            assert etcd_mod.watch(f"{prefix}/1") is None

        with subtests.test("Test no profile and bad connection in ls_"):
            assert etcd_mod.ls_() is None

        with subtests.test("Test no profile and bad connection in rm"):
            assert etcd_mod.rm_(f"{prefix}/1") is None

        with subtests.test("Test no profile and bad connection in tree"):
            assert etcd_mod.tree() is None



import logging
import os

import pytest

import salt._logging.impl as log_impl
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skip_on_windows(reason="Temporarily skipped on the newer golden images")
]


log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {log_impl: {}}


def log_nameToLevel(name):
    """
    Return the numeric representation of textual logging level
    """
    # log level values
    CRITICAL = 50
    FATAL = CRITICAL
    ERROR = 40
    WARNING = 30
    WARN = WARNING
    INFO = 20
    DEBUG = 10
    NOTSET = 0

    _nameToLevel = {
        "CRITICAL": CRITICAL,
        "FATAL": FATAL,
        "ERROR": ERROR,
        "WARN": WARNING,
        "WARNING": WARNING,
        "INFO": INFO,
        "DEBUG": DEBUG,
        "NOTSET": NOTSET,
    }
    return _nameToLevel.get(name, None)


def test_lowest_log_level():
    ret = log_impl.get_lowest_log_level()
    assert ret is not None

    log_impl.set_lowest_log_level(log_nameToLevel("DEBUG"))
    ret = log_impl.get_lowest_log_level()
    assert ret is log_nameToLevel("DEBUG")

    log_impl.set_lowest_log_level(log_nameToLevel("WARNING"))
    ret = log_impl.get_lowest_log_level()
    assert ret is log_nameToLevel("WARNING")

    opts = {"log_level": "ERROR", "log_level_logfile": "INFO"}
    log_impl.set_lowest_log_level_by_opts(opts)
    ret = log_impl.get_lowest_log_level()
    assert ret is log_nameToLevel("INFO")


def test_get_logging_level_from_string(caplog):
    ret = log_impl.get_logging_level_from_string(None)
    assert ret is log_nameToLevel("WARNING")

    ret = log_impl.get_logging_level_from_string(log_nameToLevel("DEBUG"))
    assert ret is log_nameToLevel("DEBUG")

    ret = log_impl.get_logging_level_from_string("CRITICAL")
    assert ret is log_nameToLevel("CRITICAL")

    caplog.clear()
    with caplog.at_level(logging.WARNING):
        msg = "Could not translate the logging level string 'BADLEVEL' into an actual logging level integer. Returning 'logging.ERROR'."
        ret = log_impl.get_logging_level_from_string("BADLEVEL")
        assert ret is log_nameToLevel("ERROR")
        assert msg in caplog.text


def test_logfile_handler(caplog):
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        ret = log_impl.is_logfile_handler_configured()
        assert ret is False

        msg = "log_path setting is set to `None`. Nothing else to do"
        log_path = None
        assert log_impl.setup_logfile_handler(log_path) is None
        assert msg in caplog.text


def test_in_mainprocess():
    ret = log_impl.in_mainprocess()
    assert ret is True

    curr_pid = os.getpid()
    with patch(
        "os.getpid", MagicMock(side_effect=[AttributeError, curr_pid, curr_pid])
    ):
        ret = log_impl.in_mainprocess()
        assert ret is True
